import numpy as np
import os
import pandas as pd

__all__ = ['mu_gas', 'mu_graham', 'mu_herning']


def _mu(df, formula, temp, cas, full):
    """
    Helper for the `mu_gas` function to determine gas viscosity.

    Parameters
    ----------
    df : dataframe
        Dataframe from inorganic or organic data
    formula : str
        Molecular formula for the gas
    temp : float
        Gas temperature
    cas : str
        CAS number
    full : bool
        Flag to print more information

    Returns
    -------
    mu_gas : float
        Viscosity of gas [micropoise]
    mu_gas, cas, tmin, tmax, a, b, c, d : tuple
        Additional values returned only if full=True.
    """

    if cas:
        # new dataframe based only on row for cas number
        df = df[df['CAS No.'] == str(cas)]

    cas = df.loc[formula]['CAS No.']
    tmin = df.loc[formula]['temperature, Tmin (K)']
    tmax = df.loc[formula]['temperature, Tmax (K)']
    a = df.loc[formula]['A']
    b = df.loc[formula]['B']
    c = df.loc[formula]['C']
    d = df.loc[formula]['D']

    if temp < tmin or temp > tmax:
        raise ValueError('Temperature out of range. Applicable values are '
                         f'{tmin} - {tmax} K for {formula} gas.')

    mu = a + b * temp + c * (temp**2) + d * (temp**3)

    if full:
        return mu, cas, tmin, tmax, a, b, c, d
    else:
        return mu


def mu_gas(formula, temp, cas=None, full=False):
    """
    Viscosity of gas as a function of temperature. Results calculated from
    coefficients in Yaws' Critical Property Data for Chemical Engineers and
    Chemists [1]_. CAS (Chemical Abstracts Service) number may be required for
    some species.

    Parameters
    ----------
    formula : str
        Molecular formula of the gas.
    temp : float
        Temperature of the gas [K]
    cas : str, optional
        CAS number of the gas, required for some species [-]
    full : bool, optional
        When set to :code:`False` (default) just gas viscosity is returned.
        When set to :code:`True` then return gas viscosity and other
        information.

    Returns
    -------
    mu_gas : float
        Viscosity of gas [micropoise]

    mu_gas, cas, tmin, tmax, a, b, c, d : tuple
        Additional values are only returned when keyword :code:`full=True`.

        | mu_gas - Viscosity of gas [micropoise]
        | cas - CAS number [-]
        | tmin, tmax - Temperature range at which results are applicable [K]
        | a, b, c, d - Values for regression coefficients [-]

    Raises
    ------
    ValueError
        If gas formula is not found in csv data file.
    ValueError
        If gas temperature is not in range between tmin and tmax.

    Examples
    --------
    >>> mu_gas('CH4', 810)
    234.21

    >>> mu_gas('C2Cl2F4', 900, cas='374-07-2')
    314.90

    >>> mu_gas('H2', 404)
    113.18

    >>> mu_gas('N2', 773)
    363.82

    >>> mu_gas('N2', 773, full=True)
    (363.82, '7727-37-9', 63.15, 1970.0, 4.46, 0.63, -0.00026, 5.41e-08)

    References
    ----------
    .. [1] Carl L. Yaws. Viscosity Gas Tables 80 and 81 in Yaws' Critical
       Property Data for Chemical Engineers and Chemists. Published by Knovel,
       2014.
    """
    abs_path = os.path.dirname(os.path.abspath(__file__))

    data_inorganic = abs_path + '/data/mu-gas-inorganic.csv'
    df_inorganic = pd.read_csv(data_inorganic, index_col=0)

    data_organic = abs_path + '/data/mu-gas-organic.csv'
    df_organic = pd.read_csv(data_organic, index_col=0)

    if formula in df_inorganic.index:

        if isinstance(df_inorganic.loc[formula], pd.DataFrame) and cas is None:
            raise ValueError(f'Multiple substances available for {formula}. '
                             'Include CAS number with input parameters.')

        mu = _mu(df_inorganic, formula, temp, cas, full)
        return mu

    elif formula in df_organic.index:

        if isinstance(df_organic.loc[formula], pd.DataFrame) and cas is None:
            raise ValueError(f'Multiple substances available for {formula}. '
                             'Include CAS number with input parameters.')

        mu = _mu(df_organic, formula, temp, cas, full)
        return mu

    else:
        raise ValueError(f'Gas viscosity for {formula} is not available.')


def mu_graham(mus, xs):
    """
    Calculate viscosity of a gas mixture using Graham's method [2]_. Formula
    presented here is based on Equation 1 from the Davidson report [3]_.

    .. math:: \\mu_{mix} = \\sum (x_i \\cdot \\mu_i)

    where :math:`\\mu_{mix}` is viscosity of the gas mixture, :math:`x_i` is
    mole fraction [-] of each component, and :math:`\\mu_i` is gas viscosity of
    each component.

    Parameters
    ----------
    mus : list, tuple, or array
        Viscosity of each gas component.
    xs : list, tuple, or array
        Mole fraction of each gas component [-]

    Returns
    -------
    mu : float
        Gas viscosity of the mixture. Units are same as input viscosity.

    Raises
    ------
    ValueError
        If sum of mole fractions does not equal 1.0

    Example
    -------
    >>> mu_h2 = cm.mu_gas('H2', 773.15)
    ... mu_n2 = cm.mu_gas('N2', 773.15)
    ... mu_graham([mu_h2, mu_n2], [0.85, 0.15])
    207.37

    References
    ----------
    .. [2] Thomas Graham. On the Motion of Gases. Philosophical Transactions of
       the Royal Society of London, vol. 136, pp. 573-631, 1846.
    .. [3] Thomas Davidson. A Simple and Accurate Method for Calculating
       Viscosity of Gaseous Mixtures. United States Department of the Interior:
       Report of Investigations 9456, 1993.
    """
    mus = np.asarray(mus)
    xs = np.asarray(xs)
    if not np.isclose(xs.sum(), 1.0):
        raise ValueError('Sum of mole fractions must be 1.0')
    mu_mix = np.sum(mus * xs)
    return mu_mix


def mu_herning(mus, mws, xs):
    """
    Calculate viscosity of a gas mixture using the approach by Herning and
    Zipperer [4]_. Formula presented here is based on Equation 1 from the
    Davidson report [5]_.

    .. math:: \\mu_{mix} = \\frac{\\sum (\\mu_i \\cdot x_i \\cdot \\sqrt{MW_i})}{\\sum (x_i \\cdot \\sqrt{MW_i})}

    where :math:`\\mu_{mix}` is viscosity of the gas mixture, :math:`x_i` is
    mole fraction [-] of each component, :math:`\\mu_i` is gas viscosity of
    each component, and :math:`MW_i` is the molecular weight [g/mol] of each
    component.

    Parameters
    ----------
    mus : list, tuple, or array
        Viscosity of each gas component.
    mws : list, tuple, or array
        Molecular weight of each gas component [g/mol]
    xs : list, tuple, or array
        Mole fraction of each gas component [-]

    Returns
    -------
    mu : float
        Gas viscosity of the mixture. Units are same as input viscosity.

    Raises
    ------
    ValueError
        If sum of mole fractions does not equal 1.0

    Example
    -------
    >>> mu_h2 = cm.mu_gas('H2', 773.15)
    ... mu_n2 = cm.mu_gas('N2', 773.15)
    ... mw_h2 = cm.mw('H2')
    ... mw_n2 = cm.mw('N2')
    ... mu_herning([mu_h2, mu_n2], [mw_h2, mw_n2], [0.85, 0.15])
    252.81

    References
    ----------
    .. [4] F. Herning and L. Zipperer. Calculation of the Viscosity of
       Technical Gas Mixtures From the Viscosity of the Individual Gases.
       Gas-und Wasserfach, vol. 79, pp. 69-73, 1936.
    .. [5] Thomas Davidson. A Simple and Accurate Method for Calculating
       Viscosity of Gaseous Mixtures. United States Department of the Interior:
       Report of Investigations 9456, 1993.
    """
    mus = np.asarray(mus)
    mws = np.asarray(mws)
    xs = np.asarray(xs)
    if not np.isclose(xs.sum(), 1.0):
        raise ValueError('Sum of mole fractions must be 1.0')
    mu = np.sum(mus * xs * np.sqrt(mws)) / np.sum(xs * np.sqrt(mws))
    return mu




def cp_wood(x, tk):
    """
    Heat capacity of wood based on moisture content and temperature

    .. math:: c_{p,x} = \\left(c_{p0} + c_{pw} \\frac{x}{100}\\right) / \\left(1 + \\frac{x}{100}\\right) + A_c

    where :math:`c_{p,x}` is heat capacity of wet wood [kJ/(kg K)],
    :math:`c_{p0}` is heat capacity of dry wood [kJ/(kg K)], :math:`c_{pw}` is
    heat capacity of water as 4.18 kJ/(kg K), :math:`x` is moisture content [%],
    and :math:`Ac` is an adjustment factor that accounts for the additional
    energy in the wood–water bond [1]_.

    The :math:`c_{p0}` term is determined from

    .. math:: c_{p0} = 0.1031 + 0.003867\\,T

    where :math:`T` is temperature [K]. The :math:`A_c` term is calculated from

    .. math:: A_c = x (b_1 + b_2 T + b_3 x)

    with :math:`b_1 = -0.06191`, :math:`b_2 = 2.36e\\times10^{-4}`, and :math:`b_3 = -1.33\\times10^{-4}`.

    Parameters
    ----------
    x : float
        Moisture content [%]
    tk : float
        Temperature [K]

    Returns
    -------
    cp : float
        Heat capacity of wood [kJ/(kg K)]

    Example
    -------
    >>> cp_wood(12, 340)
    1.91

    References
    ----------
    .. [1] Samuel V. Glass and Samuel L. Zelinka. Moisture Relations and
       Physical Properties of Wood. Ch. 4 in Wood Handbook, pp. 1-19, 2010.
    """

    cpw = 4.18  # heat capacity of water, kJ/(kg K)

    # coefficients for adjustment factor Ac
    b1 = -0.06191
    b2 = 2.36e-4
    b3 = -1.33e-4

    # adjustment factor for additional energy in wood-water bond, Eq. 4-18
    Ac = x * (b1 + b2 * tk + b3 * x)

    # heat capacity of dry wood, Eq. 4-16a, kJ/(kg K)
    cp_dry = 0.1031 + 0.003867 * tk

    # heat capacity of wood that contains water, Eq. 4-17, kJ/(kg K)
    cp_wet = (cp_dry + cpw * x / 100) / (1 + x / 100) + Ac

    return cp_wet



import chemics as cm
import pandas as pd
import re
from collections import Counter


class ChemicalEquation:
    """
    Properties of the reactants and products in a given chemical equation.

    Parameters
    ----------
    eq : str
        Chemical equation such as A + B -> C + D
    names : dict, optional
        Names of unique species and their corresponding chemical formula.

    Attributes
    ----------
    eq : str
        Chemical equation such as A + B -> C + D
    names : dict, optional
        Names of unique species and their corresponding chemical formula.
    rct_properties : dataframe
        Number of moles, chemical species, molecular weight, mass, mole
        fraction, and mass fraction for each reactant.
    rct_elements : dict
        Total number of atomic elements on reactant side of the equation.
    rct_moles : float
        Total number of moles on reactant side of the equation.
    rct_mass : float
        Total mass of the reactants.
    prod_properties : dataframe
        Number of moles, chemical species, molecular weight, mass, mole
        fraction, and mass fraction for each product.
    prod_elements : dict
        Total number of atomic elements on product side of the equation.
    prod_moles : float
        Total number of moles on product side of the equation.
    prod_mass : float
        Total mass of the products.
    balance : bool
        Balance between atomic elements in reactants and products.

    Examples
    --------
    >>> ce = ChemicalEquation('2 HCl + 2 Na -> 2 NaCl + H2')
    >>> ce.balance
    True

    >>> ce.rct_properties
                   HCl        Na
    moles            2         2
    species        HCl        Na
    molwt       36.458     22.99
    mass        72.916     45.98
    molfrac        0.5       0.5
    massfrac  0.613275  0.386725

    >>> ce.rct_properties.loc['massfrac']
    HCl    0.613275
    Na     0.386725

    >>> ce.rct_elements
    {'H': 2.0, 'Cl': 2.0, 'Na': 2.0}

    >>> ce.rct_moles
    4.0

    >>> ce.rct_mass
    118.896
    """

    def __init__(self, eq, names=None):
        self.eq = eq
        self.names = names
        self._parse_equation()
        self._assign_rct_attrs()
        self._assign_prod_attrs()
        self._check_balance()

    def _parse_equation(self):
        """
        Split chemical equation into reactants and products.
        """
        rct_split = self.eq.split(' -> ')[0].split(' + ')
        self._rct_items = rct_split

        prod_split = self.eq.split(' -> ')[1].split(' + ')
        self._prod_items = prod_split

    def _eq_properties(self, eq_items):
        """
        Determine properties for each item in reactants or products.
        """
        eq_names = []
        eq_moles = []
        eq_species = []
        eq_molwts = []
        eq_masses = []
        eq_elements = Counter()

        for item in eq_items:

            # number of moles and name for each item
            if item[0].isdigit():
                item = item.split()
                mol = float(item[0])
                name = item[1]
                eq_names.append(name)
                eq_moles.append(mol)
            else:
                mol = 1.0
                name = item
                eq_names.append(name)
                eq_moles.append(mol)

            # species for each item is chemical formula specified in names
            # dictonary or as chemical formula (name) in the chemical equation
            # note - species is used for the molecular weight and elements
            if (self.names is not None) and (name in self.names.keys()):
                sp = self.names[name]
                eq_species.append(sp)
            else:
                sp = name
                eq_species.append(sp)

            # masses and molecular weights for each item
            mw = cm.mw(sp)
            mass = mol * mw
            eq_molwts.append(cm.mw(sp))
            eq_masses.append(mass)

            # count elements from each chemical species
            rex = re.findall('([A-Z][a-z]?)([0-9]*)', sp)

            for r in rex:
                element = r[0]

                if r[1] == '':
                    atoms = 1.0
                else:
                    atoms = float(r[1])

                # dict where key is element and value is number of atoms such as {'C': 6.0}
                d = {element: atoms * mol}
                eq_elements.update(d)

        # sum of moles and masses
        eq_sum_moles = sum(eq_moles)
        eq_sum_masses = sum(eq_masses)

        # mole fractions and mass fractions
        eq_mol_fracs = [m / eq_sum_moles for m in eq_moles]
        eq_mass_fracs = [m / eq_sum_masses for m in eq_masses]

        # dataframe for properties
        cols = eq_names
        idx = ['moles', 'species', 'molwt', 'mass', 'molfrac', 'massfrac']

        df = pd.DataFrame(columns=cols, index=idx)
        df.loc['moles'] = eq_moles
        df.loc['species'] = eq_species
        df.loc['molwt'] = eq_molwts
        df.loc['mass'] = eq_masses
        df.loc['molfrac'] = eq_mol_fracs
        df.loc['massfrac'] = eq_mass_fracs

        return df, dict(eq_elements), eq_sum_moles, eq_sum_masses

    def _assign_rct_attrs(self):
        """
        Assign results for reactants to attributes.
        """
        df, elements, sum_moles, sum_masses = self._eq_properties(self._rct_items)
        self.rct_properties = df
        self.rct_elements = elements
        self.rct_moles = sum_moles
        self.rct_mass = sum_masses

    def _assign_prod_attrs(self):
        """
        Assign results for products to attributes.
        """
        df, elements, sum_moles, sum_masses = self._eq_properties(self._prod_items)
        self.prod_properties = df
        self.prod_elements = elements
        self.prod_moles = sum_moles
        self.prod_mass = sum_masses

    def _check_balance(self):
        """
        Check balance of atomic elements in reactants or products.
        """
        tol = 1e-4
        self.balance = True

        if self.rct_elements == self.prod_elements:
            return

        for x, y in zip(self.rct_elements.values(), self.prod_elements.values()):
            if abs(x - y) > tol:
                self.balance = False
                break



# flake8: noqa

from .atomic_elements import atomic_elements

from .bed_expansion_factor import fbexp

from .biomass_composition import biocomp
from .biomass_composition import plot_biocomp

from .bubble_rise_velocity import ubr_kunii
from .bubble_rise_velocity import ubr_holland

from .chemical_equation import ChemicalEquation

from .choking_velocity import uch_bifan
from .choking_velocity import uch_leung
from .choking_velocity import uch_matsen
from .choking_velocity import uch_psri
from .choking_velocity import uch_punwani
from .choking_velocity import uch_yang
from .choking_velocity import uch_yousfi
from .choking_velocity import uch_zhang

from .conversions import slm_to_lpm
from .conversions import massfrac_to_molefrac
from .conversions import molefrac_to_massfrac

from .devol_time import devol_time

from .dimensionless_numbers import archimedes
from .dimensionless_numbers import biot
from .dimensionless_numbers import prandtl
from .dimensionless_numbers import pyroI
from .dimensionless_numbers import pyroII
from .dimensionless_numbers import reynolds

from .gas_density import rhog
from .gas_heat_capacity import cp_gas
from .gas_pressure import patm
from .gas_thermal_conductivity import k_gas_inorganic
from .gas_thermal_conductivity import k_gas_organic
from .gas_viscosity import mu_gas
from .gas_viscosity import mu_graham
from .gas_viscosity import mu_herning

from .geldart import geldart_chart

from .minimum_fluidization_velocity import umf_coeff
from .minimum_fluidization_velocity import umf_ergun
from .minimum_fluidization_velocity import umf_reynolds

from .molecular_weight import mw
from .molecular_weight import mw_mix

from .proximate_bases import proximate_bases

from .terminal_velocity import ut
from .terminal_velocity import ut_haider
from .terminal_velocity import ut_ganser

from .transport_disengaging_height import tdh_chan
from .transport_disengaging_height import tdh_horio
from .transport_velocity import utr

from .ultimate_bases import ultimate_bases

from .wood_heat_capacity import cp_wood
from .wood_thermal_conductivity import k_wood



import numpy as np


def tdh_chan(ug):
    """
    Calculate transport disengaging height (TDH) based on the Chan and Knowlton
    correlation [1]_. This function is based on the equation presented in the
    Cahyadi [2]_ review article

    .. math:: TDH = 0.85\\,{U_g}^{1.2} (7.33 - 1.2 \\log U_g)

    where :math:`U_g` is superficial gas velocity. According to Table 1 in the
    Cahyadi review paper, this corrleation is relevant for Geldart A and B sand
    particles with a density of 2595 kg/m³, mean particle size of 37-420 µm,
    and an inner diameter fluidizing column of 0.3 m.

    Parameters
    ----------
    ug : float
        Superficial gas velocity [m/s]

    Returns
    -------
    tdh : float
        Transport disengaging height [m]

    Example
    -------
    >>> tdh_chan(0.3)
    1.7587

    References
    ----------
    .. [1] I.H. Chan and T.M. Knowlton. The effect of system pressure on the
       transport disengaging height (TDH) above bubbling gas-fluidized beds.
       AIChE Symp. Ser., vol. 80, no. 241, 1984.
    .. [2] Andy Cahyadi, Anthony H. Neumayer, Christine M. Hrenya, Ray A. Cocco,
       and Jia Wei Chew. Comparative study of Transport Disengaging Height (TDH)
       correlations in gas–solid fluidization. Powder Technology, vol. 275,
       pp. 220-238, 2015.
    """
    tdh = 0.85 * (ug**1.2) * (7.33 - 1.2 * np.log(ug))
    return tdh


def tdh_horio(dc, ug):
    """
    Calculate transport disengaging height (TDH) based on the Horio et al.
    correlation [3]_. This function is based on the equation presented in the
    Cahyadi [4]_ review article

    .. math:: TDH = (2.7\\, {D_c}^{-0.36} - 0.7) D_c\\, \\textrm{exp}(0.74\\, U_g\\, D_c^{-0.23})

    where :math:`D_c` is inner diameter of the fluidizing column and
    :math:`U_g` is superifical gas velocity. According to Table 1 in the
    Cahyadi review paper, this corrleation is relevant for Geldart group A
    particles.

    Parameters
    ----------
    dc : float
        Inner diameter of fluidizing column [m]
    ug : float
        Superficial gas velocity [m/s]

    Returns
    -------
    tdh : float
        Transport disengaging height [m]

    Example
    -------
    >>> tdh_horio(0.05, 0.3)
    0.563

    References
    ----------
    .. [3] M. Horio, T. Shibata, and I. Muchi. Design criteria for the fluidized
       bed freeboard. 4th International Conference on Fluidization in Kashikojima,
       Japan, 1983.
    .. [4] Andy Cahyadi, Anthony H. Neumayer, Christine M. Hrenya, Ray A. Cocco,
       and Jia Wei Chew. Comparative study of Transport Disengaging Height (TDH)
       correlations in gas–solid fluidization. Powder Technology, vol. 275,
       pp. 220-238, 2015.
    """
    tdh = ((2.7 * dc**-0.36) - 0.7) * dc * np.exp(0.74 * ug * dc**-0.23)
    return tdh



import pandas as pd
from pathlib import Path


def cp_gas(formula, T, CAS='', full=False):
    """
    Calculate heat capacity of a gas as a function of temperature using
    coefficients from Yaws' Critical Property Data for Chemical Engineers and
    Chemists [1]_. The CAS (Chemical Abstracts Service) number may be required for
    some species.

    .. math:: C_p = A + B\\,T + C\\,T^2 + D\\,T^3 + E\\,T^4 + F\\,T^5 + G\\,T^6

    Parameters
    ----------
    formula : str
        Chemical formula.
    T : float
        Temperature [K].
    CAS : str, optional
        CAS number for the given chemical formula.
    full : bool, optional
        When `False` (default) only gas heat capacity is returned, otherwise
        heat capacity and other information are returned.

    Returns
    -------
    Cp : float
        Heat capacity [J/(mol⋅K)].

    If `full` is `True` then also return

    CAS : str
        CAS number for the given chemical formula.
    Tmin : float
        Minimum temperature for the correlation [K].
    Tmax : float
        Maximum temperature for the correlation [K].
    A, B, C, D, E, F, G : float
        Coefficients used to calculate the heat capacity.

    Raises
    ------
    ValueError
        If heat capacity is not available for chemical formula.
    ValueError
        If multiple substances have same chemical then CAS number is required.
    ValueError
        If temperature is outside range of applicable data.

    Examples
    --------
    >>> cp_gas('CBrClF2', 700)
    97.4982

    >>> cp_gas('C5H10O2', 850, CAS='75-98-9')
    268.4920

    >>> cp_gas('NO2', 900)
    51.0686

    References
    ----------
    .. [1] Carl L. Yaws. Heat capacity of gas Tables 39 and 40 in Yaws'
       Critical Property Data for Chemical Engineers and Chemists. Published
       by Knovel, 2014.
    """
    abs_path = Path(__file__).parent.absolute()

    path_org = abs_path / 'data/cp-gas-organic.csv'
    path_inorg = abs_path / 'data/cp-gas-inorganic.csv'

    df_org = pd.read_csv(path_org, index_col=0)
    df_inorg = pd.read_csv(path_inorg, index_col=0)

    # Determine if formula is organic or inorganic species otherwise it's not
    # available in the Yaws' correlation data.
    if formula in df_org.index:
        df = df_org
    elif formula in df_inorg.index:
        df = df_inorg
    else:
        raise ValueError(f'Heat capacity for {formula} is not available.')

    # Require CAS number if multiple species are found.
    if isinstance(df.loc[formula], pd.DataFrame) and CAS == '':
        raise ValueError(f'Multiple substances available for {formula}. '
                         'Include CAS number with input parameters.')
    elif isinstance(df.loc[formula], pd.DataFrame):
        df = df[df['CAS No.'] == CAS]

    ser = df.loc[formula]

    CAS = ser['CAS No.']
    Tmin = ser['temperature, Tmin (K)']
    Tmax = ser['temperature, Tmax (K)']
    A = ser['A']
    B = ser['B']
    C = ser['C']
    D = ser['D']
    E = ser['E']
    F = ser['F']
    G = ser['G']

    if T < Tmin or T > Tmax:
        raise ValueError('Temperature out of range. Applicable values are '
                         f'{Tmin}-{Tmax} K for {formula} gas.')

    Cp = A + B * T + C * T**2 + D * T**3 + E * T**4 + F * T**5 + G * T**6

    if full:
        return Cp, CAS, Tmin, Tmax, A, B, C, D, E, F, G
    else:
        return Cp


if __name__ == '__main__':

    cp1 = cp_gas('CBrClF2', 700)
    print('cp1', cp1)

    # cp2 = cp_gas('C5H10O2', 850)
    # print('cp2', cp2)

    cp2 = cp_gas('C5H10O2', 850, CAS='75-98-9')
    print('cp2', cp2)

    cp3 = cp_gas('NO2', 900)
    print('cp3', cp3)



import matplotlib.pyplot as plt
from matplotlib.ticker import FormatStrFormatter
import numpy as np


def geldart_chart(dp, rhog, rhos, dpmin=None, dpmax=None):
    """
    Creata a Matplotlib figure of the Geldart chart [1]_. Data points for
    drawing the demarcation lines and regions were digitized from Figure 2 in
    [2]_.

    Parameters
    ----------
    dp : float
        Sauter mean particle diameter [µm]
    rhog : float
        Gas density [g/cm³]
    rhos : float
        Particle density [g/cm³]
    dpmin : float, optional
        Minimum particle size from particle size distribution [µm]
    dpmax : float, optional
        Max particle size from particle size distribution [µm]

    Returns
    -------
    Matplotlib figure.

    Note
    ----
    Particle diameter must be in microns (µm) and density in g/cm³.

    Examples
    --------
    For a single particle size

    >>> geldart_figure(300, 0.1, 2.5)

    For a particle size with min and max sizes

    >>> geldart_figure(300, 0.1, 2.5, 100, 500)

    References
    ----------
    .. [1] D. Geldart. Types of gas fluidization. Powder Technology,
       7, pp. 285-292, 1973.
    .. [2] H.A. Khawaja. Review of the phenomenon of fluidization
       and its numerical modelling techniques. The International Journal of
       Multiphysics, vol. 9, no. 4, pp. 397-407, 2015.
    """

    # Data points that form demarcation lines and regions on the Geldart chart.
    # Points are digitized from Figure 2 in Khawaja 2015.
    acx1 = [20.58, 20.77, 20.96, 20.79, 20.98, 21.17, 21.54, 22.01, 22.78,
            23.57, 24.80, 26.09, 27.80, 29.87, 32.23, 34.92, 38.16, 41.52,
            45.55, 49.98, 54.83, 60.41, 66.84, 73.95, 81.82, 86.43]

    acy1 = [6.60, 5.46, 4.51, 3.68, 3.05, 2.53, 2.11, 1.78, 1.52, 1.29, 1.12,
            0.97, 0.85, 0.76, 0.67, 0.61, 0.55, 0.49, 0.45, 0.41, 0.37, 0.34,
            0.31, 0.29, 0.27, 0.25]

    acx2 = [18.00, 17.86, 17.87, 17.88, 18.05, 18.36, 18.84, 19.33, 20.01,
            20.87, 21.96, 23.30, 24.93, 27.02, 29.27, 31.58, 34.36, 37.39,
            41.02, 45.01, 49.38, 54.64, 60.45, 66.88, 74.00, 77.84]

    acy2 = [5.63, 4.58, 3.76, 3.12, 2.57, 2.16, 1.82, 1.53, 1.31, 1.12, 0.97,
            0.86, 0.76, 0.68, 0.60, 0.54, 0.49, 0.44, 0.40, 0.36, 0.33, 0.30,
            0.28, 0.26, 0.24, 0.23]

    aax = [64.18, 70.99, 78.52, 85.77, 91.00, 94.54, 96.18, 96.23, 95.86,
           94.31, 92.78, 90.89, 87.18, 81.53, 74.66, 67.79, 61.80, 56.81,
           52.44, 49.64, 48.18, 47.36, 46.74, 46.92, 47.09, 47.87, 49.69,
           53.36, 58.03, 64.18]

    aay = [1.22, 1.22, 1.22, 1.18, 1.10, 1.00, 0.91, 0.83, 0.75, 0.67, 0.61,
           0.55, 0.51, 0.47, 0.45, 0.44, 0.45, 0.46, 0.48, 0.52, 0.57, 0.63,
           0.71, 0.79, 0.88, 0.98, 1.07, 1.14, 1.19, 1.22]

    # Combine data points into single line to fill shaded region
    acx = acx1 + acx2[::-1] + [acx1[0]]
    acy = acy1 + acy2[::-1] + [acy1[0]]

    # Equation 6 from Geldart 1973
    d_ab = np.linspace(36, 1068)
    rho_ab = 225 / d_ab

    # Equation 8 from Geldart 1973
    d_bd = np.linspace(380, 2228)
    rho_bd = (10**6) / (d_bd**2)

    # Density difference
    rhodiff = rhos - rhog

    # Figure representing Geldart chart
    fig, ax = plt.subplots(tight_layout=True)

    if dpmin and dpmax is not None:
        low_err = abs(dp - dpmin)
        upper_err = abs(dp - dpmax)
        err = [[low_err], [upper_err]]
        ax.errorbar(dp, rhodiff, fmt='ro', xerr=err, capsize=3)
    else:
        ax.plot(dp, rhodiff, 'ro')

    ax.loglog(d_ab, rho_ab, color='lightsteelblue')
    ax.loglog(d_bd, rho_bd, color='lightsteelblue')
    ax.fill_between(acx, acy, facecolor='lightsteelblue')
    ax.fill_between(aax, aay, facecolor='lightsteelblue')
    ax.text(20, 0.3, 'C\ncohesive', horizontalalignment='center')
    ax.text(68, 0.7, "A'", horizontalalignment='center')
    ax.text(200, 0.4, 'A\naeratable', horizontalalignment='center')
    ax.text(200, 3, 'B\nsand-like', horizontalalignment='center')
    ax.text(1400, 3, 'D\nspoutable', horizontalalignment='center')
    ax.set_xlabel('d$_{\\mathrm{p}}$ [µm]')
    ax.set_ylabel(r'$\mathrm{\rho_s - \rho_g}$ [g/cm³]')
    ax.set_axisbelow(True)
    ax.set_frame_on(False)
    ax.set_xlim([10, 3000])
    ax.set_xticks([10, 30, 100, 300, 1000, 3000])
    ax.set_ylim([0.1, 10])
    ax.set_yticks([0.1, 0.3, 1, 3, 10])
    ax.grid(color='0.9')
    ax.grid(which='minor', color='0.9')
    ax.tick_params(color='0.7')
    ax.tick_params(which='minor', color='0.7')
    ax.xaxis.set_major_formatter(FormatStrFormatter('%g'))
    ax.yaxis.set_major_formatter(FormatStrFormatter('%g'))

    return fig



import numpy as np


def patm(alt):
    """
    Determine atmospheric pressure at altitudes up to 51 km. Based on article
    by Roland Stull [#stull]_.

    Parameters
    ----------
    alt : float
        Altitude or elevation above sea level [m]

    Returns
    -------
    Patm : float
        Atmospheric pressure [Pa]

    References
    ----------
    .. [#stull] Practical Meteorology: An Algebra-based Survey of Atmospheric
       Science by Roland Stull.
    """
    alt = alt/1000              # convert altitude from meters to kilometers
    Ro = 6356.766               # average radius of the Earth, km
    H = (Ro * alt)/(Ro + alt)   # geopotential height, km

    Patm = None     # initiate pressure variable

    if H <= 11:
        T = 288.15 - (6.5 * H)
        Patm = 101325 * (288.15 / T) ** (-5.255877)
    elif H <= 20:
        T = 216.65
        Patm = 22632 * np.exp(-0.1577 * (H - 11))
    elif H <= 32:
        T = 216.65 + (H - 20)
        Patm = 5474.9 * (216.65 / T) ** (34.16319)
    elif H <= 47:
        T = 228.65 + 2.8 * (H - 32)
        Patm = 868 * (228.65 / T) ** 12.2011
    elif H <= 51:
        T = 270.65
        Patm = 110.9 * np.exp(-0.1262 * (H - 47))
    else:
        raise ValueError('geopotential height must be less than 51 km')

    # return atmospheric pressure at altitude, Pa
    return Patm



import numpy as np


def uch_bifan(ar, dp, gs, rhog):
    """
    Choking velocity from Bi and Fan [1]_, also see Zhang 2015 paper.

    Parameters
    ----------
    ar : float
        Archimedes number [-]
    dp : float
        Diameter of particle [m]
    gs : float
        Solids flux [kg/(s m^2)]
    rhog : float
        Density of gas [kg/m^3]

    Returns
    -------
    uch : float
        Choking velocity [m/s]

    References
    ----------
    .. [1] Bi, Fan, 1991. Regime transitions in gas-solid circulating fluidized
       beds. AIChE Annual Meeting. AIChE, Los Angeles, pp. 17-22, 1991.
    """
    g = 9.81    # gravity constant, m/s^2
    uch = (21.6 * np.sqrt(g * dp) * (gs**0.542) / (rhog**0.542) * ar**0.105)**(1 / (1 + 0.542))
    return uch


def uch_leung(us, ut):
    """
    Choking velocity based on Equation 6 from Leung [2]_. All parameters must
    be in units of ft/s or in units of m/s.

    Parameters
    ----------
    us : float
        Solids velocity [ft/s or m/s]
    ut : float
        Terminal velocity of a single particle [ft/s or m/s]

    Returns
    -------
    uch : float
        Choking velocity [ft/s or m/s]

    References
    ----------
    .. [2] L.S. Leung, Robert J. Wiles, and Donald J. Nicklin. Correlation for
       Predicting Choking Flowrates in Vertical Pneumatic Conveying. Ind. Eng.
       Chem. Process Des. Develop., vol. 10, no. 2, pp. 183-189, 1971.
    """
    uch = 32.3 * us + 0.97 * ut
    return uch


def uch_matsen(gs, rhos, ut):
    """
    Choking velocity from the Matsen paper [3]_.

    Parameters
    ----------
    gs : float
        Solids flux [kg/(s m^2)]
    rhos : float
        Density of particle [kg/m^3]
    ut : float
        Terminal velocity [m/s]

    Returns
    -------
    uch : float
        Choking velocity [m/s]

    References
    ----------
    .. [3] Matsen, 1982. Mechanism of Choking and Entrainment. Powder
       Technology, 32, 21-33.
    """
    uch = 10.74 * ut * (gs / rhos)**0.227
    return uch


def uch_psri(x, dp, d_pipe, gs, rhog, rhos, ut):
    """
    Choking velocity from PSRI 2016 notebook. Uses MKS units for input
    parameters but returns Uch in ft/s. Function requires SciPy fsolve to solve
    for Uch.

    Parameters
    ----------
    x : float
        Solve for Uch with SciPy fsolve
    dp : float
        Particle size [m]
    d_pipe : float
        Pipe diameter [m]
    gs : float
        Solids mass flux [kg/(s m^2)]
    rhog : float
        Density of gas [kg/m^3]
    rhos : float
        Density of particle [kg/m^3]
    ut : float
        Terminal velocity [m/s]

    Returns
    -------
    uch : float
        Choking velocity [ft/s]

    References
    ----------
    Notebook from PSRI 2016 workshop, pages J-20 and J-72.
    """
    uch = x                     # define variable to solve for
    g = 32.2                    # gravitational constant, ft/s^2
    dp = dp * 3.28084           # convert from m to ft
    d_pipe = d_pipe * 3.28084   # convert from m to ft
    gs = gs * 0.204817303       # convert from kg/(s m^2) to lb/(s ft^2)
    rhog = rhog * 0.062428      # convert from kg/m^3 to lb/ft^3
    rhos = rhos * 0.062428      # convert from kg/m^3 to lb/ft^3
    ut = ut * 3.28084           # convert from m/s to ft/s
    f1 = (uch - ut) / ((g * dp)**0.5) - ((gs / (uch * rhog))**0.35 * (d_pipe / dp)**0.35 * (rhos / rhog)**0.1)
    return f1


def uch_punwani(xy, d_pipe, gs, rhog, rhos, ut):
    """
    Provide functions to solve for voidage and choking velocity based on
    equations from Punwani 1976. Requires SciPy fsolve function.

    Parameters
    ----------
    xy
        Solves for ep and uch using SciPy fsolve where
        ep = voidage [-] and uch = choking velocity [m/s]
    d_pipe : float
        Internal pipe diameter [m]
    gs : float
        Solids flux [kg/(s m^2)]
    rhog : float
        Density of gas, must be in units of lb/ft^3
    rhos : float
        Density of particle [kg/m^3]
    ut : float
        Terminal velocity [m/s]

    Returns
    -------
    f1, f2 : functions
        Functions to solve for ep and uch

    References
    ----------
    Punwani, Modi, Tarman, 1976. A Generalized Correlation for Estimating
    Choking Velocity in Vertical Solids Transport. Institute of Gas Technology,
    Chicago, IL, 1-14.
    """
    ep, uc = xy     # define variables
    g = 9.81        # accelerationg of gravity, m/s^2

    f1 = 2 * g * d_pipe * ((ep**-4.7) - 1) - (0.074 * rhog**0.77) * (uc / ep - ut)**2
    f2 = (gs / rhos) / (1 - ep) - (uc / ep) + ut
    return f1, f2


def uch_yang(xy, d_pipe, gs, rhos, ut):
    """
    Provide functions to solve for voidage and choking velocity based on
    equations from Yang 1975. Requires SciPy fsolve function.

    Parameters
    ----------
    xy
        Solve for ep and uch using SciPy fsolve where
        ep = voidage [-] and uch = choking velocity [m/s]
    d_pipe : float
        Internal pipe diameter [m]
    gs : float
        Solids flux [kg/(s m^2)]
    rhos : float
        Density of particle [kg/m^3]
    ut : float
        Terminal velocity [m/s]

    Returns
    -------
    f1, f2 : fuctions
        Functions to solve for ep and uch

    References
    ----------
    Yang, 1975. A Mathematical Definition of Choking Phenomenon and a
    Mathematical Model for Predicting Choking Velocity and Choking Voidage.
    AIChE Journal, 21, 5, 1013-1015.
    """
    ep, uch = xy     # define variables
    g = 9.81        # acceleration due to gravity, m/s^2

    f1 = 2 * g * d_pipe * ((ep**-4.7) - 1) - 0.01 * (uch / ep - ut)**2
    f2 = gs - (uch / ep - ut) * rhos * (1 - ep)
    return f1, f2


def uch_yousfi(dp, gs, rhog, mu, ut):
    """
    Choking velocity from Yousfi 1974, also see Zhang 2015 paper.

    Parameters
    ----------
    dp : float
        Particle diameter [ft or m]
    gs : float
        Solids flux [lb/(hr ft^2) or kg/(s m^2)]
    rhog : float
        Gas density [lb/ft^3 or kg/m^3]
    mu : float
        Gas viscosity [lb/(ft s) or kg/(s m)]
    ut : float
        Terminal velocity of particle [ft/s or m/s]

    Returns
    -------
    uch : float
        Choking velocity [ft/s or m/s]

    References
    ----------
    Yousfi, Gau, 1974. Aerodynamique De L'Ecoulement Verical De Suspensions
    Concentrees Gaz-Solides I. Regimes D'Ecoulement Et Stabilite Aerodynamique.
    Chemical Engineering Science, 29, 1939-1946.
    """
    g = 9.81                    # accelaration due to gravity, m/s^2
    re = (dp * rhog * ut) / mu      # Reynolds number at terminal velocity
    uch = (32 * np.sqrt(g * dp) * (re**-0.06) * (gs**0.28 / rhog**0.28))**(1 / (1 + 0.28))
    return uch


def uch_zhang(ar, dp, gs, rhog):
    """
    Choking velocity from Zhang 2015.

    Parameters
    ----------
    ar : float
        Archimedes number [-]
    dp : float
        Diameter of particle [m]
    gs : float
        Solids flux [kg/(s m^2)]
    rhog : float
        Density of gas [kg/m^3]

    Returns
    -------
    uch : float
        Choking velocity [m/s]

    Reference
    ---------
    Zhang, Degreve, Dewil, Baeyens, 2015. Operation diagram of Circulating
    Fluidized Beds (CFBs). Procedia Engineering, 102, 1092-1103.
    """
    g = 9.81    # gravity constant, m/s^2
    uch = (14.6 * np.sqrt(g * dp) * (gs**0.542) / (rhog**0.542) * ar**0.105)**(1 / (1 + 0.542))
    return uch



import numpy as np


def ubr_holland(db, rho_l, rho_g, sig, mu_l):
    """
    Bubble rise velocity in a liquid for a range of applicability. This
    velocity is related to the terminal rise velocity correlations for a
    single bubble in a liquid as given in Table 7.1 in the Holland and Bragg
    book [1]_.

    Parameters
    ----------
    db : float
        Diameter of sphere having same volume as spherical cap bubble,
        referred to as the effective bubble diameter [m]
    rho_l : float
        Density of the liquid [kg/m^3]
    rho_g : float
        Density of the gas [kg/m^3]
    sig : float
        Surface tension [dynes/cm]
    mu_l : float
        Dynamic viscosity of the liquid [centiPoise]

    Returns
    -------
    ub : float
        Bubble (terminal) rise velocity [m/s]

    Example
    -------
    >>> ubr_holland(0.00003, 998, 0, 72.75, 1.21)
    0.0004046

    References
    ----------
    .. [1] F.A. Holland and R. Bragg. Fluid Flow for Chemical Engineers.
           Butterworth-Heinemann, 2nd edition, p. 234, 1995.
    """

    # Constants
    g = 9.81    # acceleration due to gravity [m/s^2]

    # Convert all given parameters to mks units
    r_b_mks = db / 2        # [m]
    sig_mks = sig / 1000    # [N/m]
    mu_l_mks = mu_l / 1000  # [Pa s]

    # bubble Reynolds number [-]
    def Re(ub):
        return 2 * rho_l * ub * r_b_mks / mu_l_mks

    # Morton Number [-]
    G1 = (g * mu_l_mks**4) / (rho_l * sig_mks**3)

    # Region 1: Bubbles behave as bouyant spheres, rising vertically.
    ub = (2 * (r_b_mks**2) * (rho_l - rho_g) * g) / (9 * mu_l_mks)

    if Re(ub) <= 2:
        return ub

    # Region 2: Bubbles raise as spheres, but the drag coefficient slightly
    #           less than that of a solids of the same volume.
    ub = 0.33 * (g**0.76) * ((rho_l / mu_l_mks)**0.52) * (r_b_mks**1.28)

    if Re(ub) > 2 and Re(ub) <= 4.02 * (G1**-0.214):
        return ub

    # Region 3: Bubbles are flattened and rise in zig-zag pattern.
    ub = 1.35 * (sig_mks / (rho_l * r_b_mks))**0.5

    if Re(ub) > 4.02 * (G1**-0.214) and Re(ub) <= 3.10 * (G1**-0.25):
        return ub

    # Region 4: Bubbles rise vertically adopting a mushroom-cap shape.
    ub = 1.18 * ((sig * g / rho_l)**0.25)

    if Re(ub) > 3.10 * G1**-0.25 and Re(ub) <= 2.3 * (sig / (g * rho_l))**0.5:
        return ub

    # Region 5: Large spherical-cap bubbles.
    ub = (g * r_b_mks)**0.5

    return ub


def ubr_kunii(db, dt):
    """
    Rise velocity of single bubbles in a fluidized bed from Equations 5.3 and
    5.4 in Kunii and Levenspiel book [2]_.

    Parameters
    ----------
    db : float
        Diameter of sphere having same volume as spherical cap bubble,
        referred to as the effective bubble diameter [m]
    dt : float
        Bed or tube diameter [m]

    Returns
    -------
    ubr : float
        Rise velocity of single bubbles in fluidized bed [m/s]

    Example
    -------
    >>> ubr_kunii(0.05, 0.6)
    0.4979

    References
    ----------
    .. [2] Daizo Kunii and Octave Levenspiel. Fluidization Engineering.
           Butterworth-Heinemann, 2nd edition, 1991.
    """
    g = 9.81    # acceleration of gravity [m/s^2]

    if db / dt < 0.125:
        ubr = 0.711 * (g * db)**(1 / 2)
    elif 0.125 <= db / dt < 0.6:
        ubr = (0.711 * (g * db)**(1 / 2)) * 1.2 * np.exp(-1.49 * db / dt)
    else:
        raise ValueError('Bed is considered slugging at db/dt > 0.6')

    return ubr



import numpy as np


def devol_time(dp, tbed):
    """
    Correlation for 95% devolatilization time for wood particles. Refer to
    Equation 2 in the Di Blasi [1]_ article which is based on beech wood
    cylinders in a fluidized sand bed.

    Parameters
    ----------
    dp : float or array
        Diameter of biomass particle [mm]
    tbed : float or array
        Bed temperature [K]

    Returns
    -------
    tv : float or array
        Devolatilization time for 95% conversion [s]

    Example
    -------
    >>> devol_time(2.0, 773.15)
    13.2114

    References
    ----------
    .. [1] Colomba Di Blasi and Carmen Branca. Temperatures of Wood Particles
       in a Hot Sand Bed Fluidized by Nitrogen. Energy and Fuels, 17,
       pp. 247-254, 2003.
    """
    tv = 0.8 * np.exp(1525 / tbed) * (dp**1.2)
    return tv



import numpy as np


def ut(cd, dp, rhog, rhos):
    """
    Calculate terminal velocity of a single particle based on Equation 28 on
    page 80 in the Kunii and Levenspiel book [1]_ where :math:`C_D` is an
    experimentally determined drag coefficient.

    .. math:: u_t = \\left( \\frac{4 d_p\\, (\\rho_s - \\rho_g) g}{3 \\rho_g\\, C_D} \\right)^{1/2}

    Parameters
    ----------
    cd : float
        Drag coefficient [-]
    dp : float
        Diameter of particle [m]
    rhog : float
        Density of gas [kg/m^3]
    rhos : float
        Density of solid [kg/m^3]

    Returns
    -------
    ut : float
        Terminal velocity [m/s]

    Example
    -------
    >>> ut(11.6867, 0.00016, 1.2, 2600)
    0.6227

    References
    ----------
    .. [1] Daizo Kunii and Octave Levenspiel. Fluidization Engineering.
           Butterworth-Heinemann, 2nd edition, 1991.
    """
    g = 9.81  # gravity acceleration, m/s^2
    tm1 = 4 * dp * (rhos - rhog) * g
    tm2 = 3 * rhog * cd
    ut = (tm1 / tm2)**(1 / 2)
    return ut


def ut_ganser(dp, mu, phi, rhog, rhos):
    """
    Estimate terminal velocity of a non-spherical particle based on the Ganser
    drag coefficient [2]_. According to the Chhabra paper [3]_, the Ganser drag
    correlation is applicable for sphericity values from 0.09 to 1.

    .. math::

       C_d &= \\frac{24}{Re\\, K_1} \\left( 1 + 0.1118 (Re\\, K_1 K_2)^{0.6567} \\right) + \\frac{0.4305 K_2}{1 + \\frac{3305}{Re\\, K_1 K_2}}

       K_1 &= \\left( \\frac{1}{3} + \\frac{2}{3}\\phi \\right)

       K_2 &= 10^{1.8148 (-log\\, \\phi)^{0.5743}}

    where K₁ is Stokes' shape factor and K₂ is Newton's shape factor. The Cui 2007
    and Chhabra 1999 papers leave out the :math:`-2.25*d_v/D` term in the shape
    factor equations.

    Parameters
    ----------
    dp : float
        Diameter of the particle [m]
    mu : float
        Viscosity of gas [kg/(m s)]
    phi : float
        Sphericity of the particle [-]
    rhog : float
        Density of the gas [kg/m³]
    rhos : float
        Density of the particle [kg/m³]

    Returns
    -------
    ut : float
        Terminal velocity of non-spherical particle [m/s]

    Example
    -------
    >>> ut_ganser(0.00016, 1.8e-5, 0.67, 1.2, 2600)
    0.6230

    Note
    ----
    Drag coefficient is referenced from Equation 18 in Ganser, Equation 6 in
    Chhabra, and as Equation 2 in Cui [4]_.

    References
    ----------
    .. [2] Gary H. Ganser. A rational approach to drag prediction of spherical
           and nonsperical particles. Powder Technology, 77, 143-152, 1993.
    .. [3] R.P. Chhabra, L. Agarwal, and N.K. Sinha. Drag on non-spherical
           particles: An evaluation of available methods. Powder Technology,
           101, 288-295, 1999.
    .. [4] Heping Cui and John R. Grace. Fluidization of biomass particles: A
           review of experimental multiphase flow aspects. Chemical Engineering
           Science, 62, 45-55, 2007.
    """
    g = 9.81    # acceleration from gravity [m/s²]

    # Guess a range for Ut [m/s] to perform calculations
    # max terminal velocity in range determined from Newton's law
    ut_newton = 1.74 * np.sqrt(9.81 * dp * (rhos - rhog) / rhog)
    ut = np.arange(0.0001, ut_newton, 0.004)

    # Shape factors
    # papers Cui 2007 and Chhabra 1999 leave out the -2.25*dv/D term
    k1 = (1 / 3 + 2 / 3 * (phi**-0.5))**(-1)           # Stokes' shape factor
    k2 = 10**(1.8148 * ((-np.log(phi))**0.5743))       # Newton's shape factor

    # Evaluate Ganser drag coefficient [-] for a range of Ut
    # re and cd are vectors
    re = (dp * rhog * ut) / mu
    cd = (24 / (re * k1)) * (1 + 0.1118 * ((re * k1 * k2)**0.6567)) \
        + (0.4305 * k2) / (1 + (3305 / (re * k1 * k2)))

    # Evaluate sphere drag coefficient [-] for a range of Ut
    # this expression is from the Levenspiel book
    # cdd is a vector
    cdd = (4 * g * dp * (rhos - rhog)) / (3 * (ut**2) * rhog)

    # Interpolate value for Ut where Cd-Cdd = 0
    ut_interp = np.interp(0, cd - cdd, ut)

    return ut_interp


def ut_haider(dp, mu, phi, rhog, rhos):
    """
    Calculate terminal velocity of a particle as discussed in the Haider and
    Levenspiel [5]_. Valid for particle sphericities of 0.5 to 1. Particle
    diameter should be an equivalent spherical diameter, such as the diameter
    of a sphere having same volume as the particle.

    To determine the terminal velocity for a range of particle sphericities, Haider
    and Levenspiel first define two dimensionless quantities

    .. math::

       d_{*} = d_p \\left[ \\frac{g\\, \\rho_g (\\rho_s - \\rho_g)}{\\mu^2} \\right]^{1/3} \\
       u_* = \\left[ \\frac{18}{d{_*}^2} + \\frac{2.3348 - 1.7439\\, \\phi}{d{_*}^{0.5}} \\right]^{-1}

    where :math:`0.5 \\leq \\phi \\leq 1` and particle diameter :math:`d_p` is an
    equivalent spherical diameter, the diameter of a sphere having the same volume
    as the particle. The relationship between :math:`u_*` and :math:`u_t` is given
    by

    .. math:: u_* = u_t \\left[ \\frac{\\rho{_g}^2}{g\\, \\mu\\, (\\rho_s - \\rho_g)} \\right]^{1/3}

    The terminal velocity of the particle can finally be determined by rearranging
    the above equation such that

    .. math:: u_t = u_* \\left[ \\frac{g\\, \\mu\\, (\\rho_s - \\rho_g)}{\\rho{_g}^2} \\right]^{1/3}

    Parameters
    ----------
    dp : float
        Diameter of particle [m]
    mu : float
        Viscosity of gas [kg/(m s)]
    phi : float
        Sphericity of particle [-]
    rhog : float
        Density of gas [kg/m^3]
    rhos : float
        Density of particle [kg/m^3]

    Returns
    -------
    ut : float
        Terminal velocity of a particle [m/s]

    Example
    -------
    >>> ut_haider(0.00016, 1.8e-5, 0.67, 1.2, 2600)
    0.8857

    References
    ----------
    .. [5] A. Haider and O. Levenspiel. Drag coefficient and terminal velocity
           of spherical and nonspherical particles. Powder Technology,
           58:63–70, 1989.
    """
    if phi > 1.0 or phi < 0.5:
        raise ValueError('Sphericity must be 0.5 <= phi <= 1.0')

    d_star = dp * ((9.81 * rhog * (rhos - rhog)) / (mu**2))**(1 / 3)
    u_star = (18 / (d_star**2) + ((2.3348 - 1.7439 * phi) / (d_star**0.5)))**-1
    ut = u_star * ((9.81 * (rhos - rhog) * mu) / rhog**2)**(1 / 3)
    return ut



import numpy as np


def eexp(db, hb, rhop, mp):
    """
    Calculates the overall void fraction of the expanded bed.

    Parameters
    ----------
    db : float
        Diameter of the bed [m]
    hb : float
        Height of the expanded bed [m]
    rhop : float
        Mass density of the solids [kg/m^3]
    mp : float
        Mass of solids in the bed [kg]

    Note
    ----
    Specific mass and length units are not important, provided they are
    all self-consistent as the returned quantity is dimensionless.

    Returns
    -------
    ep_exp : float
        Overall void fraction of the expanded bed [-]
    """

    # Volume of the expanded bed
    vb = np.pi * (db**2 / 4.0) * hb

    # Volume due to the solids
    vp = mp / rhop

    # Overall void fraction
    ep_exp = 1.0 - float(vp) / vb

    return ep_exp


def emf(fbexp, ep_exp):
    """
    Calculate the void fraction at minimum fluidization for the specified bed.

    Parameters
    ----------
    fbexp : float
        Bed expansion factor [-]
    ep_exp : float
        Void fraction in the expanded bed [-]

    Returns
    -------
    e_mf : float
        Void fraction at minimum fluidization [-]
    """

    e_mf = 1 - fbexp * (1 - ep_exp)
    return e_mf


def ueGas(ub, umf, emf, delta, fw):
    """
    This function calculates the velocity of the gas in the emulsion phase with
    K/L eqs. 6.39-6.40.

    Parameters
    ----------
    ub : float
        Bubble rise velocity [m/s]
    umf : float
        Minimum fluidization velocity [m/s]
    emf : float
        Void fraction at minimum fluidization [-]
    delta : float
        Fraction of bed volume in bubbles [-]
    fw : float
        Ratio of wake volume to bubble volume [-]

    Returns
    -------
    ue : float
        Emulsion gas velocity [m/s]
    """

    ue = umf / emf - fw * delta * ub / (1 - delta - fw * delta)
    return ue


def solidsDistribution(delta, emf, umf, ub, fw):
    """
    Calculates the distribution of solids in the bubble, cloud/wake, and
    emulsion phases using K/L eqs. 6.35-6.37.

    Parameters
    ----------
    delta : float
        Volume fraction of gas in the bubbles [-]
    emf : float
        Void fraction at minimum fluidization [-]
    umf : float
        Minimum fluidization velocity [m/s]
    ub : float
        Bubble rise velocity [m/s]
    fw : float
        Ratio of the wake volume to the bubble volume [-]

    Note
    ----
    gamma_b is for the moment a hard-coded constant. See K/L eq. 6.37.

    Returns
    -------
    gamma_b : float
        Volume of solids in the bubble divided by the bubble volume
    gamma_c : float
        Volume of solids in the cloud/wake divided by the bubble volume
    gamma_e : float
        Volume of solids in the emulsion divided by the bubble volume
    """

    gamma_b = 0.005
    gamma_c = (1.0 - emf) * (3.0/(ub * float(emf) / umf - 1.0) + fw)
    gamma_e = (1.0 - emf) * (1.0 - delta) / delta - gamma_b - gamma_c
    return gamma_b, gamma_c, gamma_e


def massTransferCoeff(umf, emf, db, ub, diff, approx=True, method=None):
    """
    Calculates the rate constant for gas exchange between the bubble and the
    emulsion. See K/L eqs. 10.27, 10.34, and 10.36.

    Parameters
    ----------
    umf : float
        Minimum fluidization velocity [m/s]
    emf : float
        Void fraction at minimum fluidization [-]
    db : float
        Bubble diameter [m]
    ub : float
        Bubble rise velocity [m/s]
    diff : float
        Diffusion coefficient [m^2/s]
    approx : boolean
        Control whether to use a simpler expression that does not employ the
        diffusion coefficient
    method : string
        Correlation to use as either KL for Kunii-Levenspiel or SG for
        Sit-Grace method

    Returns
    -------
    kbe : float
        Rate constant (s^-1) for exchanging gas between the bubble and the
        emulsion
    """

    # Constants
    g = 9.81    # m/s^2

    if method == 'KL':
        # First check the bubbling regime
        if approx and (1 < ub / (umf / emf) < 5):
            # Intermediate bubbling, use a simpler approximation that doesn't
            # require the (poorly known) diffusion coefficient. See K/L eq.
            # 12.47.
            kbe = 4.5 * umf / db
            return kbe
        else:
            # We should use the full equation.

            # Bubble/cloud exchange (10.27) in units of [1/s]
            kbc = 4.5 * (umf / db) + 5.85 \
                * (np.sqrt(diff) * g**0.25 / db**1.25)

            # Cloud/emulsion exchange (10.34)
            kce = 6.77 * np.sqrt(diff * emf * ub / db**3)

            # Overall rate
            inv_kbe = 1.0/kbc + 1.0/kce
            kbe = 1.0/inv_kbe

            return kbe

    elif method == 'SG':

        kbe = 2*umf/db + 12.0/db**1.5 * np.sqrt(diff*emf*ub/np.pi)
        return kbe

    else:

        raise NotImplementedError(
            'Unknown mass transfer coefficient estimation method ' +
            str(method))


def solidsFluxWenChen(db, dt, umf, us, rhog, ug):
    """
    Calculates the solids flux at the surface of the bubbling bed according to
    the method of Wen and Chen (AIChE J, 28, 1, pp. 117-128) with a correction
    recommended by Wells, Kulver, and Krishnan in ORNL/TM-7847 (TVA AFBC
    Simulation Interim Annual Report, Dec. 1980) for large bed diameters.

    Parameters
    ----------
    db : float
        Bubble diameter [m]
    dt : float
        Bed/tank diameter [m]
    umf : float
        Minimum fluidization velocity [m/s]
    us : float
        Superficial velocity [m/s]
    rhog : float
        Gas mass density [kg/m^3]
    ug : float
        Gas viscosity [kg/m/s]

    Returns
    -------
    F0 : float
        Mass flux of entrained solids [kg/m^2/s]
    """

    g = 9.81    # m/s^2

    # Tube cross-sectional area, m^2
    ac = np.pi * dt**2 / 4

    # Solids flux
    F0 = 3.07e-9 * (ac * db * (us - umf)**2.5 * rhog**2.5 * np.sqrt(g)) \
        / (ug**2.5 * umf)

    # Correction for small aspect ratio beds (wider than it is deep)
    if float(db) / dt < 1:
        F0 *= (float(db) / dt)**2

    return F0


def solidsFluxPembNose(dp, db, emf, umf, us):
    """
    Calculates the solids entrainment flux for the surface of a bubbling bed
    with the bubble nose ejection model of Pemberton and Davidsion (Chem. Eng.
    Sci., 1986, 41, pp. 243-251). This model is suitable if there is a single
    bubble in the bed.

    Parameters
    ----------
    dp : float
        Particle diameter [m]
    db : float
        Bubble diameter [m]
    emf : float
        Void fraction at minimum fluidization [-]
    umf : float
        Minimum fluidization velocity [m/s]
    us : float
        Superficial gas velocity [m/s]

    Returns
    -------
    F0 : float
        Solids entrainment flux at the top of the bed [kg/m^2/s]
    """

    F0 = 3.0 * dp / db * (1 - emf) * (us - umf)
    return F0


def solidsFluxPembWake(rhop, emf, umf, us):
    """
    Calculates the solids entrainment flux for the surface of a bubbling bed
    with the bubble wake ejection model of Pemberton and Davidsion (Chem. Eng.
    Sci., 1986, 41, pp. 243-251). This model is suitable if multiple bubbles
    coalesce at the surface before bursting.

    Parameters
    ----------
    rhop : float
        Mass density of the particles
    emf : float
        Void fraction at minimum fluidization [-]
    umf : float
        Minimum fluidization velocity [m/s]
    us : float
        Superficial gas velocity [m/s]

    Returns
    -------
    F0 : float
        Solids entrainment flux at the top of the bed [kg/m^2/s]
    """

    F0 = 0.1 * rhop * (1 - emf) * (us - umf)
    return F0



import numpy as np
import textwrap


def proximate_bases(fc, vm, ash, moisture, disp=False):
    """
    Convert proximate analysis from as-received basis (% ar) to dry basis (%
    dry) and dry-ash free basis (% daf).

    Parameters
    ----------
    fc : float
        Percent fixed carbon
    vm : float
        Percent volatile matter
    ash : float
        Percent ash
    moisture : float
        Percent moisture
    disp : bool, optional
        Print results to console, default is `False`

    Returns
    -------
    bases : dict
        Proximate analysis bases calculated from given as-received basis.
        Bases are represented by keys in the dictionary where `'ar'` is
        as-received basis with values of `[FC, VM, ash, moisture]`, `'dry'` is
        dry basis with values of `[FC, VM, ash]`, and `'daf'` is dry ash-free
        basis with values of `[FC, VM]`.

    Raises
    ------
    ValueError
        If the proximate analysis sum is not 100.

    Example
    -------
    >>> proximate_bases(16.92, 76.40, 0.64, 6.04)
    {
        'ar': [16.92, 76.4, 0.64, 6.04],
        'dry': [18.00, 81.31, 0.68],
        'daf': [18.13, 81.86]
    }
    """

    # as-received basis (% ar)
    prox_ar = [fc, vm, ash, moisture]
    sum_ar = sum(prox_ar)

    # make sure proximate analysis sums to 100
    if not np.isclose(sum_ar, 100.0):
        raise ValueError('Sum of proximate analysis values must be 100')

    # dry basis (% dry)
    prox_dry = [100 * x / (sum_ar - prox_ar[-1]) for x in prox_ar[:-1]]
    sum_dry = sum(prox_dry)

    # dry ash-free basis (% daf)
    prox_daf = [100 * x / (sum_dry - prox_dry[-1]) for x in prox_dry[:-1]]
    sum_daf = sum(prox_daf)

    # print results to console if `disp=True`
    if disp:
        results = textwrap.dedent(f"""
                        % ar    % dry    % daf
        FC          {prox_ar[0]:8} {prox_dry[0]:8.2f} {prox_daf[0]:8.2f}
        VM          {prox_ar[1]:8} {prox_dry[1]:8.2f} {prox_daf[1]:8.2f}
        ash         {prox_ar[2]:8} {prox_dry[2]:8.2f}
        moisture    {prox_ar[3]:8}
        sum         {sum_ar:8.2f} {sum_dry:8.2f} {sum_daf:8.2f}
        """)
        print(results)

    # results dictionary
    bases = {
        'ar': prox_ar,
        'dry': prox_dry,
        'daf': prox_daf
    }

    return bases



"""
Dictionary containing atomic number, name, and atomic weight of elements in the
periodic table. Conventional atomic weight is used for atomic weight where
applicable. Values taken from IUPAC.

References
----------
IUPAC Periodic Table of the Elements. International Union of Pure and Applied
Chemistry, 2016. https://iupac.org/what-we-do/periodic-table-of-elements/.
"""

atomic_elements = {
    'H': {'atomic_number': 1, 'name': 'hydrogen', 'atomic_weight': 1.008},
    'He': {'atomic_number': 2, 'name': 'helium', 'atomic_weight': 4.0026},
    'Li': {'atomic_number': 3, 'name': 'lithium', 'atomic_weight': 6.94},
    'Be': {'atomic_number': 4, 'name': 'beryllium', 'atomic_weight': 9.0122},
    'B': {'atomic_number': 5, 'name': 'boron', 'atomic_weight': 10.81},
    'C': {'atomic_number': 6, 'name': 'carbon', 'atomic_weight': 12.011},
    'N': {'atomic_number': 7, 'name': 'nitrogen', 'atomic_weight': 14.007},
    'O': {'atomic_number': 8, 'name': 'oxygen', 'atomic_weight': 15.999},
    'F': {'atomic_number': 9, 'name': 'fluorine', 'atomic_weight': 18.998},
    'Ne': {'atomic_number': 10, 'name': 'neon', 'atomic_weight': 20.180},
    'Na': {'atomic_number': 11, 'name': 'sodium', 'atomic_weight': 22.990},
    'Mg': {'atomic_number': 12, 'name': 'magnesium', 'atomic_weight': 24.305},
    'Al': {'atomic_number': 13, 'name': 'aluminium', 'atomic_weight': 26.982},
    'Si': {'atomic_number': 14, 'name': 'silicon', 'atomic_weight': 28.085},
    'P': {'atomic_number': 15, 'name': 'phosphorus', 'atomic_weight': 30.974},
    'S': {'atomic_number': 16, 'name': 'sulfur', 'atomic_weight': 32.06},
    'Cl': {'atomic_number': 17, 'name': 'chlorine', 'atomic_weight': 35.45},
    'Ar': {'atomic_number': 18, 'name': 'argon', 'atomic_weight': 39.948},
    'K': {'atomic_number': 19, 'name': 'potassium', 'atomic_weight': 39.098},
    'Ca': {'atomic_number': 20, 'name': 'calcium', 'atomic_weight': 40.078},
    'Sc': {'atomic_number': 21, 'name': 'scandium', 'atomic_weight': 44.956},
    'Ti': {'atomic_number': 22, 'name': 'titanium', 'atomic_weight': 47.867},
    'V': {'atomic_number': 23, 'name': 'vanadium', 'atomic_weight': 50.942},
    'Cr': {'atomic_number': 24, 'name': 'chromium', 'atomic_weight': 51.996},
    'Mn': {'atomic_number': 25, 'name': 'manganese', 'atomic_weight': 54.938},
    'Fe': {'atomic_number': 26, 'name': 'iron', 'atomic_weight': 55.845},
    'Co': {'atomic_number': 27, 'name': 'cobalt', 'atomic_weight': 58.933},
    'Ni': {'atomic_number': 28, 'name': 'nickel', 'atomic_weight': 58.693},
    'Cu': {'atomic_number': 29, 'name': 'copper', 'atomic_weight': 63.546},
    'Zn': {'atomic_number': 30, 'name': 'zinc', 'atomic_weight': 65.38},
    'Ga': {'atomic_number': 31, 'name': 'gallium', 'atomic_weight': 69.723},
    'Ge': {'atomic_number': 32, 'name': 'germanium', 'atomic_weight': 72.630},
    'As': {'atomic_number': 33, 'name': 'arsenic', 'atomic_weight': 74.922},
    'Se': {'atomic_number': 34, 'name': 'selenium', 'atomic_weight': 78.971},
    'Br': {'atomic_number': 35, 'name': 'bromine', 'atomic_weight': 79.904},
    'Kr': {'atomic_number': 36, 'name': 'krypton', 'atomic_weight': 83.798},
    'Rb': {'atomic_number': 37, 'name': 'rubidium', 'atomic_weight': 85.468},
    'Sr': {'atomic_number': 38, 'name': 'strontium', 'atomic_weight': 87.62},
    'Y': {'atomic_number': 39, 'name': 'yttrium', 'atomic_weight': 88.906},
    'Zr': {'atomic_number': 40, 'name': 'zirconium', 'atomic_weight': 91.224},
    'Nb': {'atomic_number': 41, 'name': 'niobium', 'atomic_weight': 92.906},
    'Mo': {'atomic_number': 42, 'name': 'molybdenum', 'atomic_weight': 95.95},
    'Tc': {'atomic_number': 43, 'name': 'technetium', 'atomic_weight': None},
    'Ru': {'atomic_number': 44, 'name': 'ruthenium', 'atomic_weight': 101.07},
    'Rh': {'atomic_number': 45, 'name': 'rhodium', 'atomic_weight': 102.91},
    'Pd': {'atomic_number': 46, 'name': 'palladium', 'atomic_weight': 106.42},
    'Ag': {'atomic_number': 47, 'name': 'silver', 'atomic_weight': 107.87},
    'Cd': {'atomic_number': 48, 'name': 'cadmium', 'atomic_weight': 112.41},
    'In': {'atomic_number': 49, 'name': 'indium', 'atomic_weight': 114.82},
    'Sn': {'atomic_number': 50, 'name': 'tin', 'atomic_weight': 118.71},
    'Sb': {'atomic_number': 51, 'name': 'antimony', 'atomic_weight': 121.76},
    'Te': {'atomic_number': 52, 'name': 'tellurium', 'atomic_weight': 127.60},
    'I': {'atomic_number': 53, 'name': 'iodine', 'atomic_weight': 126.90},
    'Xe': {'atomic_number': 54, 'name': 'xenon', 'atomic_weight': 131.29},
    'Cs': {'atomic_number': 55, 'name': 'caesium', 'atomic_weight': 132.91},
    'Ba': {'atomic_number': 56, 'name': 'barium', 'atomic_weight': 137.33},
    'La': {'atomic_number': 57, 'name': 'lanthanum', 'atomic_weight': 138.91},
    'Ce': {'atomic_number': 58, 'name': 'cerium', 'atomic_weight': 140.12},
    'Pr': {'atomic_number': 59, 'name': 'praseodymium', 'atomic_weight': 140.91},
    'Nd': {'atomic_number': 60, 'name': 'neodymium', 'atomic_weight': 144.24},
    'Pm': {'atomic_number': 61, 'name': 'promethium', 'atomic_weight': None},
    'Sm': {'atomic_number': 62, 'name': 'samarium', 'atomic_weight': 150.36},
    'Eu': {'atomic_number': 63, 'name': 'europium', 'atomic_weight': 151.96},
    'Gd': {'atomic_number': 64, 'name': 'gadolinium', 'atomic_weight': 157.25},
    'Tb': {'atomic_number': 65, 'name': 'terbium', 'atomic_weight': 158.93},
    'Dy': {'atomic_number': 66, 'name': 'dysprosium', 'atomic_weight': 162.50},
    'Ho': {'atomic_number': 67, 'name': 'holmium', 'atomic_weight': 164.93},
    'Er': {'atomic_number': 68, 'name': 'erbium', 'atomic_weight': 167.26},
    'Tm': {'atomic_number': 69, 'name': 'thulium', 'atomic_weight': 168.93},
    'Yb': {'atomic_number': 70, 'name': 'ytterbium', 'atomic_weight': 173.05},
    'Lu': {'atomic_number': 71, 'name': 'lutetium', 'atomic_weight': 174.97},
    'Hf': {'atomic_number': 72, 'name': 'hafnium', 'atomic_weight': 178.49},
    'Ta': {'atomic_number': 73, 'name': 'tantalum', 'atomic_weight': 180.95},
    'W': {'atomic_number': 74, 'name': 'tungsten', 'atomic_weight': 183.84},
    'Re': {'atomic_number': 75, 'name': 'rhenium', 'atomic_weight': 186.21},
    'Os': {'atomic_number': 76, 'name': 'osmium', 'atomic_weight': 190.23},
    'Ir': {'atomic_number': 77, 'name': 'iridium', 'atomic_weight': 192.22},
    'Pt': {'atomic_number': 78, 'name': 'platinum', 'atomic_weight': 195.08},
    'Au': {'atomic_number': 79, 'name': 'gold', 'atomic_weight': 196.97},
    'Hg': {'atomic_number': 80, 'name': 'mercury', 'atomic_weight': 200.59},
    'Tl': {'atomic_number': 81, 'name': 'thallium', 'atomic_weight': 204.38},
    'Pb': {'atomic_number': 82, 'name': 'lead', 'atomic_weight': 207.2},
    'Bi': {'atomic_number': 83, 'name': 'bismuth', 'atomic_weight': 208.98},
    'Po': {'atomic_number': 84, 'name': 'polonium', 'atomic_weight': None},
    'At': {'atomic_number': 85, 'name': 'astatine', 'atomic_weight': None},
    'Rn': {'atomic_number': 86, 'name': 'radon', 'atomic_weight': None},
    'Fr': {'atomic_number': 87, 'name': 'francium', 'atomic_weight': None},
    'Ra': {'atomic_number': 88, 'name': 'radium', 'atomic_weight': None},
    'Ac': {'atomic_number': 89, 'name': 'actinium', 'atomic_weight': None},
    'Th': {'atomic_number': 90, 'name': 'thorium', 'atomic_weight': 232.04},
    'Pa': {'atomic_number': 91, 'name': 'protactinium', 'atomic_weight': 231.04},
    'U': {'atomic_number': 92, 'name': 'uranium', 'atomic_weight': 238.03},
    'Np': {'atomic_number': 93, 'name': 'neptunium', 'atomic_weight': None},
    'Pu': {'atomic_number': 94, 'name': 'plutonium', 'atomic_weight': None},
    'Am': {'atomic_number': 95, 'name': 'americium', 'atomic_weight': None},
    'Cm': {'atomic_number': 96, 'name': 'curium', 'atomic_weight': None},
    'Bk': {'atomic_number': 97, 'name': 'berkelium', 'atomic_weight': None},
    'Cf': {'atomic_number': 98, 'name': 'californium', 'atomic_weight': None},
    'Es': {'atomic_number': 99, 'name': 'einsteinium', 'atomic_weight': None},
    'Fm': {'atomic_number': 100, 'name': 'fermium', 'atomic_weight': None},
    'Md': {'atomic_number': 101, 'name': 'mendelevium', 'atomic_weight': None},
    'No': {'atomic_number': 102, 'name': 'nobelium', 'atomic_weight': None},
    'Lr': {'atomic_number': 103, 'name': 'lawrencium', 'atomic_weight': None},
    'Rf': {'atomic_number': 104, 'name': 'rutherfordium', 'atomic_weight': None},
    'Db': {'atomic_number': 105, 'name': 'dubnium', 'atomic_weight': None},
    'Sg': {'atomic_number': 106, 'name': 'seaborgium', 'atomic_weight': None},
    'Bh': {'atomic_number': 107, 'name': 'bohrium', 'atomic_weight': None},
    'Hs': {'atomic_number': 108, 'name': 'hassium', 'atomic_weight': None},
    'Mt': {'atomic_number': 109, 'name': 'meitnerium', 'atomic_weight': None},
    'Ds': {'atomic_number': 110, 'name': 'darmstadtium', 'atomic_weight': None},
    'Rg': {'atomic_number': 111, 'name': 'roentgenium', 'atomic_weight': None},
    'Cn': {'atomic_number': 112, 'name': 'copernicium', 'atomic_weight': None},
    'Nh': {'atomic_number': 113, 'name': 'nihonium', 'atomic_weight': None},
    'Fl': {'atomic_number': 114, 'name': 'flerovium', 'atomic_weight': None},
    'Mc': {'atomic_number': 115, 'name': 'moscovium', 'atomic_weight': None},
    'Lv': {'atomic_number': 116, 'name': 'livermorium', 'atomic_weight': None},
    'Ts': {'atomic_number': 117, 'name': 'tennessine', 'atomic_weight': None},
    'Og': {'atomic_number': 118, 'name': 'oganesson', 'atomic_weight': None}
}



"""
Use the `biocomp()` function to calculate biomass composition. Use the
`plot_biocomp()` function to create a Matplotlib figure of the biomass
composition results.
"""

import numpy as np


def biocomp(yc, yh, yo=None, yh2o=0, yash=0, alpha=0.6, beta=0.8, gamma=0.8,
            delta=1, epsilon=1, printcomp=False):
    """
    Determine bimoass composition from ultimate analysis mass fractions of C,
    H, and O. Composition returned as cellulose, hemicellulose, lignins, and
    extractives based on method discussed in the Debiagi 2015 paper [1]_.

    Parameters
    ----------
    yc : float
        Mass fraction of carbon in biomass, dry ash free basis [-]
    yh : float
        Mass fraction of hydrogen in biomass, dry ash free basis [-]
    yo : float, optional
        Mass fraction of oxygen in biomass, if not given then value is
        calculated as difference, dry ash free basis [-]. Default is None.
    yh2o : float, optional
        Mass fraction of water in biomass, as received basis [-]. Default is 0.
    yash : float, optional
        Mass fraction of ash in biomass, as received basis [-]. Default is 0.
    alpha : float, optional
        Splitting parameter as molar ratio of cellulose and hemicellulose
        contained in reference mixture RM1 [-]. Default is 0.6.
    beta : float, optional
        Splitting parameter as molar ratio of lignin LIG-O and lignin LIG-C
        contained in reference mixture RM2 [-]. Default is 0.8.
    gamma : float, optional
        Splitting parameter as molar ratio of lignin LIG-H and lignin LIG-C
        contained in reference mixture RM3 [-]. Default is 0.8.
    delta : float, optional
        Splitting parameter as molar ratio of lignins (LIG-H and LIG-C) and
        extractive TGL to define reference mixture RM2 [-]. Default is 1.0.
    epsilon : float, optional
        Splitting parameter as molar ratio of lignins (LIG-O and LIG-C) and
        extractive TANN to define reference mixture RM3 [-]. Default is 1.0.
    printcomp : bool, optional
        Print composition results if True. Default is False.

    Returns
    -------
    comp : dict
        Dictionary representing reference mixtures and biomass compositions on
        the basis of mole fractions (x) and mass fractions (y).

        - `y_rm1` mass fractions [C, H, O] of reference mixture RM1
        - `y_rm2` mass fractions [C, H, O] of reference mixture RM2
        - `y_rm3` mass fractions [C, H, O] of reference mixture RM3
        - `x_daf` mole fractions [cell, hemi, ligc, ligh, ligo, tann, tgl] of
          biomass as dry ash-free basis
        - `x_wet` mole fractions [cell, hemi, ligc, ligh, ligo, tann, tgl] of
          biomass as wet basis
        - `y_daf` mass fractions [cell, hemi, ligc, ligh, ligo, tann, tgl] of
          biomass as dry ash-free basis
        - `y_wet` mass fractions [cell, hemi, ligc, ligh, ligo, tann, tgl] of
          biomass as wet basis
        - `y_wetash` mass fractions [cell, hemi, ligc, ligh, ligo, tann, tgl]
          of biomass as wet ash basis

    Raises
    ------
    ValueError
        When sum of mass fractions is not equal to one.

    Examples
    --------
    >>> yc = 0.534
    >>> yh = 0.06
    >>> bc = biocomp(yc, yh)
    >>> bc['y_daf']
    array([0.2936, 0.1594, 0.0712, 0.2934, 0.1822, 0, 0])

    >>> yc = 0.500
    >>> yh = 0.060
    >>> yo = 0.440
    >>> yash = 0.15
    >>> biocomp(yc, yh, yo, yash=yash, printcomp=True)
    basis    cell    hemi    ligc    ligh    ligo    tann    tgl
    x_daf    0.5016  0.3344  0.0328  0.0614  0.0698  0.0000  0.0000
    x_wet    0.5016  0.3344  0.0328  0.0614  0.0698  0.0000  0.0000
    y_daf    0.4275  0.2322  0.0445  0.1409  0.1549  0.0000  0.0000
    y_wet    0.4275  0.2322  0.0445  0.1409  0.1549  0.0000  0.0000
    y_wetash 0.3634  0.1974  0.0378  0.1197  0.1317  0.0000  0.0000

    References
    ----------
    .. [1] Paulo Eduardo Amaral Debiagi, Chiara Pecchi, Giancarlo Gentile,
       Alessio Frassoldati, Alberto Cuoci, Tiziano Faravelli, and Eliseo Ranzi.
       Extractives Extend the Applicability of Multistep Kinetic Scheme of Biomass
       Pyrolysis. Energy and Fuels, vol. 29, no. 10, pp. 6544-6555, 2015.
    """

    # Determine oxygen mass fraction by difference if not explicitly given
    if yo is None:
        yo = 1 - yc - yh

    # Check that mass fractions sum to one
    sumy = yc + yh + yo
    tol = 1e-4
    if abs(sumy - 1.0) > tol:
        raise ValueError('Sum of mass fractions must equal one.')

    # Cellulose, hemicellulose, and lignins as arrays of [C, H, O]
    # See Figure 1 in reference for these formulas
    cell = np.array([6, 10, 5])
    hemi = np.array([5, 8, 4])
    ligc = np.array([15, 14, 4])
    ligh = np.array([22, 28, 9])
    ligo = np.array([20, 22, 10])
    tann = np.array([15, 12, 7])
    tgl = np.array([57, 100, 7])

    # Molecular weight of cellulose, hemicellulose, lignins, and water
    mw_cell = 162.141
    mw_hemi = 132.115
    mw_ligc = 258.273
    mw_ligh = 436.457
    mw_ligo = 422.386
    mw_tann = 304.254
    mw_tgl = 897.42
    mw_h2o = 18.015

    # Solve for pseudo species
    # -----------------------------------------------------------------------

    # Reference mixture where rm = [C, H, O]
    rm1 = alpha * cell + (1 - alpha) * hemi
    rm2 = beta * delta * ligh + (1 - beta) * delta * ligc + (1 - beta * delta - (1 - beta) * delta) * tgl
    rm3 = gamma * epsilon * ligo + (1 - gamma) * epsilon * ligc + (1 - gamma * epsilon - (1 - gamma) * epsilon) * tann

    # Molecular weight of reference mixture where C, H, O given as 12, 1, 16
    mw_rm1 = sum(rm1 * [12, 1, 16])
    mw_rm2 = sum(rm2 * [12, 1, 16])
    mw_rm3 = sum(rm3 * [12, 1, 16])

    # Mass fraction of reference mixture where y_rm = [yC, yH, yO]
    y_rm1 = (rm1 * [12, 1, 16]) / mw_rm1
    y_rm2 = (rm2 * [12, 1, 16]) / mw_rm2
    y_rm3 = (rm3 * [12, 1, 16]) / mw_rm3

    # Mass fraction of pseudo species where ys = [S1, S2, S3]
    # Solve system of linear equations Ax = b
    # A is 3x3 matrix with columns = [yC, yH, yO]
    a = np.array([y_rm1, y_rm2, y_rm3]).T
    b = np.array([yc, yh, yo])
    ys = np.linalg.solve(a, b)

    # Sum of mass fractions and molecular weights as sum(ys/MW)
    sum_ymw = sum(ys / [mw_rm1, mw_rm2, mw_rm3])

    # Mole fraction of pseudo species where xs = [S1, S2, S3]
    xs = (ys / [mw_rm1, mw_rm2, mw_rm3]) / sum_ymw

    # Biomass composition
    # -----------------------------------------------------------------------

    # Mole fraction as dry ash free basis for biomass components where
    # x_daf = [cell, hemi, ligc, ligh, ligo, tann, tgl]
    x_dafcell = alpha * xs[0]
    x_dafhemi = (1 - alpha) * xs[0]
    x_dafligc = (1 - beta) * delta * xs[1] + (1 - gamma) * epsilon * xs[2]
    x_dafligh = beta * delta * xs[1]
    x_dafligo = gamma * epsilon * xs[2]
    x_daftann = (1 - gamma * epsilon - (1 - gamma) * epsilon) * xs[2]
    x_daftgl = (1 - beta * delta - (1 - beta) * delta) * xs[1]
    x_daf = np.array([x_dafcell, x_dafhemi, x_dafligc, x_dafligh, x_dafligo, x_daftann, x_daftgl])

    # Molecular weight vector for biomass components where
    # mw_comp = [mw_cell, mw_hemi, mw_ligc, mw_ligh, mw_ligo, mw_tann, mw_tgl]
    mw_comp = x_daf * [mw_cell, mw_hemi, mw_ligc, mw_ligh, mw_ligo, mw_tann, mw_tgl]

    # Average molecular weight of all biomass components
    mw_avg = sum(mw_comp)

    # Mass fraction as dry ash free basis for biomass components where
    # y_daf = [cell, hemi, ligc, ligh, ligo, tann, tgl]
    y_daf = mw_comp / mw_avg

    # Mass fraction as wet basis for biomass components where
    # y_wet = [cell, hemi, ligc, ligh, ligo, tann, tgl]
    y_wet = y_daf * (1 - yh2o)

    # Mass fraction as wet + ash basis for biomass components where
    # y_wetash = [cell, hemi, ligc, ligh, ligo, tann, tgl]
    y_wetash = y_daf * (1 - yh2o - yash)

    # Mole fraction as wet basis for biomass components where
    # x_wet = [cell, hemi, ligc, ligh, ligo, tann, tgl]
    ywet_yh2o = np.concatenate((y_wet, [yh2o]))
    sum_xmw = sum(ywet_yh2o / [mw_cell, mw_hemi, mw_ligc, mw_ligh, mw_ligo, mw_tann, mw_tgl, mw_h2o])
    x_wet = (y_wet / [mw_cell, mw_hemi, mw_ligc, mw_ligh, mw_ligo, mw_tann, mw_tgl]) / sum_xmw

    comp = {
        'y_rm1': y_rm1,
        'y_rm2': y_rm2,
        'y_rm3': y_rm3,
        'x_daf': x_daf,
        'x_wet': x_wet,
        'y_daf': y_daf,
        'y_wet': y_wet,
        'y_wetash': y_wetash
    }

    # Print biomass composition results
    if printcomp:
        print('basis\t cell\t hemi\t ligc\t ligh\t ligo\t tann\t tgl')
        print('x_daf\t', '  '.join(f'{x:.4f}' for x in x_daf))
        print('x_wet\t', '  '.join(f'{x:.4f}' for x in x_wet))
        print('y_daf\t', '  '.join(f'{x:.4f}' for x in y_daf))
        print('y_wet\t', '  '.join(f'{x:.4f}' for x in y_wet))
        print('y_wetash', '  '.join(f'{x:.4f}' for x in y_wetash))

    return comp


def plot_biocomp(ax, yc, yh, y_rm1, y_rm2, y_rm3):
    """
    Plot characterization of biomass sample and calculated reference mixtures
    as mass fractions, dry ash-free basis.

    Parameters
    ----------
    ax : Axes
        The Matplotlib axes from a figure
    yc : float
        Mass fraction of carbon in biomass, dry ash free basis [-]
    yh : float
        Mass fraction of hydrogen in biomass, dry ash free basis [-]
    y_rm1 : array
        Mass fraction of reference mixture RM1 where y_rm1 = [yC, yH, yO]
    y_rm2 : array
        Mass fraction of reference mixture RM2 where y_rm2 = [yC, yH, yO]
    y_rm3 : array
        Mass fraction of reference mixture RM3 where y_rm3 = [yC, yH, yO]

    Returns
    -------
    ax : Axes
        The Matplotlib axes for the plot figure
    """

    ax.plot(yc, yh, '^', label='biomass')

    ax.plot(0.4444, 0.0617, 'o')
    ax.annotate('cell', xy=(0.4444, 0.0617), xytext=(-10, 6), textcoords='offset points')

    ax.plot(0.4545, 0.0606, 'o')
    ax.annotate('hemi', xy=(0.4545, 0.0606), xytext=(-10, -12), textcoords='offset points')

    ax.plot(0.6977, 0.0543, 'o')
    ax.annotate('ligc', xy=(0.6977, 0.0543), xytext=(-10, -12), textcoords='offset points')

    ax.plot(0.6055, 0.0642, 'o')
    ax.annotate('ligh', xy=(0.6055, 0.0642), xytext=(-10, 6), textcoords='offset points')

    ax.plot(0.5687, 0.0521, 'o')
    ax.annotate('ligo', xy=(0.5687, 0.0521), xytext=(-10, -12), textcoords='offset points')

    ax.plot(0.5921, 0.0395, 'o')
    ax.annotate('tann', xy=(0.5921, 0.0395), xytext=(-10, -12), textcoords='offset points')

    ax.plot(0.7634, 0.1116, 'o')
    ax.annotate('tgl', xy=(0.7634, 0.1116), xytext=(-6, -12), textcoords='offset points')

    x = 0.4444, 0.5921, 0.6977, 0.7634
    y = 0.0617, 0.0395, 0.0543, 0.1116
    ax.fill(x, y, '0.8', alpha=0.5)

    ax.plot(y_rm1[0], y_rm1[1], 's', label='rm1')
    ax.plot(y_rm2[0], y_rm2[1], 's', label='rm2')
    ax.plot(y_rm3[0], y_rm3[1], 's', label='rm3')

    x = y_rm1[0], y_rm2[0], y_rm3[0], y_rm1[0]
    y = y_rm1[1], y_rm2[1], y_rm3[1], y_rm1[1]
    ax.plot(x, y, 'k:')

    ax.grid(color='0.9')
    ax.set_axisbelow(True)
    ax.set_frame_on(False)
    ax.set_xlabel('Carbon mass fraction, daf basis [-]')
    ax.set_ylabel('Hydrogen mass fraction, daf basis [-]')
    ax.tick_params(color='0.9')
    ax.legend(frameon=False)

    return ax



def archimedes(dp, rhog, rhos, mu):
    """
    Calculate the dimensionless Archimedes number.

    .. math:: Ar = \\frac{d_p^3 \\rho_g (\\rho_s - \\rho_g) g}{\\mu^2}

    Parameters
    ----------
    dp : float
        Particle diameter [m]
    rhog : float
        Gas density [kg/m³]
    rhos : float
        Solid density [kg/m³]
    mu : float
        Dynamic viscosity [kg/(m⋅s)]

    Returns
    -------
    ar : float
        Archimedes number [-]

    Example
    -------
    >>> archimedes(0.001, 910, 2500, 0.001307)
    8309.1452

    References
    ----------
    Daizo Kunii and Octave Levenspiel. Fluidization Engineering.
    Butterworth-Heinemann, 2nd edition, 1991.
    """
    g = 9.81    # gravity acceleraton [m/s²]
    ar = (dp**3 * rhog * (rhos - rhog) * g) / (mu**2)
    return ar


def biot(h, d, k):
    """
    Calculate the dimensionless Biot number.

    .. math:: Bi = \\frac{h\\, d}{k}

    Parameters
    ----------
    h : float
        Convective heat transfer coefficient [W/(m²⋅K)]
    d : float
        Characteristic length or dimension [m]
    k : float
        Thermal conductivity [W/(m⋅K)]

    Returns
    -------
    bi : float
        Biot number [-]

    Example
    -------
    >>> biot(4.63, 0.001, 3.84)
    0.0012057

    References
    ----------
    Daizo Kunii and Octave Levenspiel. Fluidization Engineering.
    Butterworth-Heinemann, 2nd edition, 1991.
    """
    bi = (h * d) / k
    return bi


def prandtl(cp=None, mu=None, k=None, nu=None, alpha=None):
    """
    Calculate the dimensionless Prandtl number for a fluid or gas.

    .. math:: Pr = \\frac{c_p \\mu}{k} = \\frac{\\nu}{\\alpha}

    Parameters
    ----------
    cp : float
        Specific heat [J/(kg⋅K)]
    mu : float
        Dynamic viscosity [kg/(m⋅s)]
    k : float
        Thermal conductivity [W/(m⋅K)]
    nu : float, optional
        Kinematic viscosity [m²/s]
    alpha : float, optional
        Thermal diffusivity [m²/s]

    Returns
    -------
    pr : float
        Prandtl number [-]

    Examples
    --------
    >>> prandtl(cp=4188, mu=0.001307, k=0.5674)
    9.647

    >>> prandtl(nu=1.5064e-5, alpha=2.1002e-5)
    0.71726

    Raises
    ------
    ValueError
        Must provide (cp, mu, k) or (nu, alpha)

    References
    ----------
    Daizo Kunii and Octave Levenspiel. Fluidization Engineering.
    Butterworth-Heinemann, 2nd edition, 1991.
    """
    if cp and mu and k:
        pr = (cp * mu) / k
    elif nu and alpha:
        pr = nu / alpha
    else:
        raise ValueError('Must provide (cp, mu, k) or (nu, alpha)')

    return pr


def pyroI(k, kr, rho, cp, r):
    """
    Calculate the pyrolysis number Py I for a biomass particle.

    .. math:: Py^I = \\frac{k}{\\rho\\,C_p\\,R^2\\,K}

    Parameters
    ----------
    k : float
        Thermal conductivity of the biomass particle [W/mK]
    kr : float
        Rate constant [1/s]
    rho : float
        Density of the biomass particle [kg/m³]
    cp : float
        Heat capacity of the biomass particle [J/kgK]
    r : float
        Radius or characteristic length of the biomass particle [m]

    Returns
    -------
    pyI : float
        Pyrolysis number Py I [-]

    Example
    -------
    >>> pyroI(k=0.12, kr=1.38556, rho=540, cp=3092.871049, r=0.0001847)
    1.5198

    References
    ----------
    D.L. Pyle and C.A. Zaror. Heat Transfer and Kinetics in the Low
    Temperature Pyrolysis of Solids. Chemical Engineering Science, vol. 39,
    no. 1, pg. 147-158, 1984.
    """
    pyI = k / (kr * rho * cp * (r**2))
    return pyI


def pyroII(h, kr, rho, cp, r):
    """
    Calculate the pyrolysis number Py II for a biomass particle.

    .. math:: Py^{II} = \\frac{h}{\\rho\\,C_p\\,R\\,K}

    Parameters
    ----------
    h : float
        Convective heat transfer coefficient [W/m²K]
    kr : float
        Rate constant [1/s]
    rho : float
        Density of the biomass particle [kg/m³]
    cp : float
        Heat capacity of the biomass particle [J/kgK]
    r : float
        Radius or characteristic length of the biomass particle [m]

    Returns
    -------
    pyII : float
        Pyrolysis number Py II [-]

    Example
    -------
    >>> pyroII(h=862.6129, kr=1.38556, rho=540, cp=3092.871049, r=0.0001847)
    2.018038

    References
    ----------
    D.L. Pyle and C.A. Zaror. Heat Transfer and Kinetics in the Low
    Temperature Pyrolysis of Solids. Chemical Engineering Science, vol. 39,
    no. 1, pg. 147-158, 1984.
    """
    pyII = h / (kr * rho * cp * r)
    return pyII


def reynolds(u, d, rho=None, mu=None, nu=None):
    """
    Calculate the dimensionless Reynolds number for a fluid or gas flow.

    .. math:: Re = \\frac{\\rho\\, u\\, d}{\\mu} = \\frac{u\\, d}{\\nu}

    Parameters
    ----------
    u : float
        Flow speed [m/s]
    d : float
        Characteristic length or dimension [m]
    rho : float, optional
        Density of the fluid or gas [kg/m³]
    mu : float, optional
        Dynamic viscosity of the fluid or gas [kg/(m⋅s)]
    nu : float, optional
        Kinematic viscosity of the fluid or gas [m²/s]

    Returns
    -------
    re : float
        Reynolds number [-]

    Examples
    --------
    >>> reynolds(2.6, 0.025, rho=910, mu=0.38)
    155.65789

    >>> reynolds(0.25, 0.102, nu=1.4e-6)
    18214.2857

    Raises
    ------
    ValueError
        Must provide (u, d, rho, mu) or (u, d, nu)

    References
    ----------
    Daizo Kunii and Octave Levenspiel. Fluidization Engineering.
    Butterworth-Heinemann, 2nd edition, 1991.
    """
    if rho and mu and not nu:
        re = (rho * u * d) / mu
    elif nu and not rho and not mu:
        re = (u * d) / nu
    else:
        raise ValueError('Must provide (u, d, rho, mu) or (u, d, nu)')

    return re



import numpy as np


def massfrac_to_molefrac(y, mw):
    """
    Convert from mass fractions to mole fractions. Calculation assumes a total
    mass of 100 g.

    .. math::

       m_i = y_i \\times 100

       x_i = \\frac{\\frac{m_i}{MW_i}}{\\sum\\frac{m_i}{MW_i}}

    where :math:`m` is mass [g], :math:`y` is mass fraction [-], :math:`x` is
    mole fraction [-], and :math:`MW` is molecular weight [g/mol] of each
    component.

    Parameters
    ----------
    y : list, tuple or array
        Mass fraction of each component [-]
    mw : list, tuple or array
        Molecular weight of each component [g/mol]

    Returns
    -------
    x : array
        Mole fractions of each component [-]

    Example
    -------
    >>> y = [0.36, 0.16, 0.20, 0.28]
    ... mw = [12.011, 1.008, 15.999, 14.007]
    ... massfrac_to_molefrac(y, mw)
    [0.136 0.718 0.057 0.09]
    """

    # convert inputs to NumPy arrays
    y = np.asarray(y)
    mw = np.asarray(mw)

    if not np.isclose(y.sum(), 1.0):
        raise ValueError('Sum of mass fractions must be 1.0')

    # mass of each component assuming 100 total grams [g]
    m = y * 100

    # mole fraction of each component [-]
    x = (m / mw) / np.sum(m / mw)

    return x


def molefrac_to_massfrac(x, mw):
    """
    Convert from mole fractions to mass fractions. Calculation assumes total
    moles is 100.

    .. math::

       n_i &= x_i \\times 100

       y_i &= \\frac{n_i\\, MW_i}{\\sum n_i\\, MW_i}

    where :math:`n` is moles [mol], :math:`x` is mole fraction [-], :math:`y`
    is mass fraction [-], and :math:`MW` is molecular weight [g/mol] of each
    component.

    Parameters
    ----------
    x : list, tuple, or array
        Mole fraction of each component [-]
    mw : list, tuple or array
        Molecular weight of each component [g/mol]

    Returns
    -------
    y : array
        Mass fraction of each component [-]

    Example
    -------
    >>> x = [0.36, 0.16, 0.20, 0.28]
    >>> mw = [12.011, 1.008, 15.999, 14.007]
    ... molefrac_to_massfrac(x, mw)
    [0.373 0.014 0.276 0.338]
    """

    # convert inputs to NumPy arrays
    x = np.asarray(x)
    mw = np.asarray(mw)

    if not np.isclose(x.sum(), 1.0):
        raise ValueError('Sum of mole fractions must be 1.0')

    # moles of each component assuming 100 total moles [mol]
    n = x * 100

    # mass fraction of each component [-]
    y = (n * mw) / np.sum(n * mw)

    return y


def slm_to_lpm(slm, pgas, tgas):
    """
    Convert volumetric gas flow from standard liters per minute (SLM or SLPM)
    to liters per minute (LPM) where STP defined as 273.25 K and 101,325 Pa.

    .. math::

       1 LPM = 1 SLPM \\times \\frac{T_{gas}}{273.15\\,K} \\times \\frac{14.696\\,psi}{P_{gas}}

    Parameters
    ----------
    slm : float
        Volumetric gas flow in standard liters per minute [SLM]
    pgas : float
        Absolute gas pressure [kPa]
    tgas : float
        Gas temperature [K]

    Returns
    -------
    lpm : float
        Volumetric gas flow in liters per minute [LPM]

    Example
    -------
    >>> slm_to_lpm(580, 150, 773)
    1108.74

    References
    ----------
    Wikipedia contributors. (2018, February 8). Standard litre per minute.
    In Wikipedia online. Retrieved from
    https://en.wikipedia.org/wiki/Standard_litre_per_minute
    """

    # equation requires gas pressure as psi so convert kPa to psi
    pgas_psi = pgas * 0.1450377
    lpm = slm * (tgas / 273.15) * (14.696 / pgas_psi)
    return lpm



def utr(dp, mu, rhog, rhos):
    """
    Determine the transport velocity of particles in a circulating fluidized
    bed riser. Based on Equation 2 in article by Zhang et al. [1]_.

    Parameters
    ----------
    dp : float
        Diameter of particle [m]
    mu : float
        Viscosity of gas [kg/(m s)]
    rhog : float
        Density of gas [kg/m^3]
    rhos : float
        Density of solid particle [kg/m^3]

    Returns
    -------
    utr : float
        Transport velocity [m/s]

    Example
    -------
    >>> utr(0.0005, 3.6e-5, 0.44, 1630)
    26.0617

    References
    ----------
    .. [1] H.L. Zhang, J. Degreve, R. Dewil, and J. Baeyens. Operation diagram
        of Circulating Fluidized Beds (CFBs). Procedia Engineering, 102,
        pp. 1092-1103, 2015.
    """
    g = 9.81                                                # gravity acceleraton [m/s^2]
    ar = ((dp**3) * rhog * (rhos - rhog) * g) / (mu**2)     # Archimedes number [-]
    utr = (mu / (dp * rhog)) * (3.23 + 0.23 * ar)           # transport velocity [m/s]
    return utr



def bubbleFraction(ub, us, umf, emf):
    """
    Determine the fraction of the bed that is in the bubble phase using
    Equations 6.26 through 6.29 from Kunii and Levenspiel [1]_.

    Parameters
    ----------
    ub : float
        Bubble rise velocity [m/s, cm/s, etc.]
    us : float
        Superficial gas velocity [m/s, cm/s, etc.]
    umf : float
        Minimum fluidization velocity [m/s, cm/s, etc.]
    emf : float
        Minimum fluidization void fraction [-]

    Returns
    -------
    delta : float
        Fraction of the total bed volume that is in bubbles [-]

    Note
    ----
    Units are not important, provided they are consistent, as the calculation
    returns a dimensionless quantity.

    References
    ----------
    .. [1] Daizo Kunii and Octave Levenspiel. Fluidization Engineering.
           Butterworth-Heinemann, 2nd edition, 1991.
    """

    # Velocity of gas in the emulsion -- used to decide which regime we are in
    ue = umf / emf

    # Ratio of velocities. This ratio determines which regime we are in.
    ub_ue_ratio = float(ub) / float(ue)

    # Calculate delta
    if ub_ue_ratio < 1:     # Slow bubbles
        delta = (us - umf) / (ub + 2*umf)
    elif ub_ue_ratio > 10:  # Very fast bubbles
        delta = us / ub
    elif ub_ue_ratio > 5:   # Fast Bubbles
        delta = (us - umf) / (ub - umf)
    elif ub_ue_ratio < 3:   # Slower intermediate bubbles
        delta = (us - umf) / (ub + umf)
    else:                   # Faster intermediate bubbles
        delta = (us - umf) / ub

    return delta



import numpy as np


def umf_coeff(dp, mu, rhog, rhos, coeff='wenyu'):
    """
    Determine minimum fluidization velocity using experimental coefficients from
    Wen and Yu, Richardson, Saxena and Vogel, Babu, Grace, and Chitester. This
    approach can be used when bed void fraction and particle sphericity are not
    known. Refer to Equation 25 and Table 4 in Chapter 3 of Kunii and Levenspiel
    [1]_.

    Parameters
    ----------
    dp : float
        Diameter of bed particle [m]
    mu : float
        Viscosity of gas [kg/(m s)]
    rhog : float
        Density of gas [kg/m³]
    rhos : float
        Density of bed particle [kg/m³]
    coeff : string
        Keyword to determine which coefficients to use for umf calculation.
        Valid options are 'wenyu', 'rich', 'sax', 'babu', 'grace', and 'chit'.
        Default coefficients are set to 'wenyu'.

    Returns
    -------
    umf : float
        Minimum fluidization velocity [m/s]

    Example
    -------
    >>> umf_coeff(0.0005, 3.6e-5, 0.44, 2500, 'rich')
    0.1192

    References
    ----------
    .. [1] Daizo Kunii and Octave Levenspiel. Fluidization Engineering.
       Butterworth-Heinemann, 2nd edition, 1991.
    """

    if coeff == 'wenyu':
        # Wen and Yu coefficients [-]
        a = 33.7
        b = 0.0408
    elif coeff == 'rich':
        # Richardson coefficients [-]
        a = 25.7
        b = 0.0365
    elif coeff == 'sax':
        # Saxena and Vogel coefficients [-]
        a = 25.3
        b = 0.0571
    elif coeff == 'babu':
        # Babu coefficients [-]
        a = 25.3
        b = 0.0651
    elif coeff == 'grace':
        # Grace coefficients [-]
        a = 27.2
        b = 0.0408
    elif coeff == 'chit':
        # Chitester coefficients [-]
        a = 28.7
        b = 0.0494
    else:
        raise ValueError('Coefficient is not a valid option.')

    # g is acceleration due to gravity [m/s²], Ar is Archimedes number [-], and
    # Re is Reynolds number [-]
    g = 9.81
    Ar = (dp**3 * rhog * (rhos - rhog) * g) / (mu**2)
    Re = (a**2 + b * Ar)**0.5 - a

    # minimum fluidization velocity [m/s]
    umf = (Re * mu) / (dp * rhog)
    return umf


def umf_ergun(dp, ep, mu, phi, rhog, rhos):
    """
    Determine minimum fluidization velocity from particle and gas properties.
    This approach is based on the Ergun pressure drop equation for a bed of
    particles. Refer to Equations 18 and 19 in Chapter 3 of Kunii and Levenspiel
    [2]_,

    Parameters
    ----------
    dp : float
        Diameter of bed particle [m]
    ep : float
        Void fraction of the bed [-]
    mu : float
        Viscosity of gas [kg/ms]
    phi : float
        Sphericity of bed particle [-]
    rhog : float
        Density of gas [kg/m³]
    rhos : float
        Density of bed particle [kg/m³]

    Returns
    -------
    umf : float
        Minimum fluidization velocity [m/s]

    Example
    -------
    >>> umf_ergun(0.0005, 0.46, 3.6e-5, 0.86, 0.44, 2500)
    0.1488

    References
    ----------
    .. [2] Daizo Kunii and Octave Levenspiel. Fluidization Engineering.
       Butterworth-Heinemann, 2nd edition, 1991.
    """

    # g is acceleration from gravity [m/s²]
    # K1, K2, a, b are dimensionless constants [-]
    g = 9.81
    K1 = 1.75 / (ep**3 * phi)
    K2 = 150 * (1 - ep) / (ep**3 * phi**2)
    a = K2 / (2 * K1)
    b = 1 / K1

    # Ar is Archimedes number [-], Re is Reynolds number [-], and umf is minimum
    # fluidization velocity [m/s]
    Ar = ((dp**3) * rhog * (rhos - rhog) * g) / (mu**2)
    Re = ((a**2 + b * Ar)**0.5) - a
    umf = (Re * mu) / (rhog * dp)
    return umf


def umf_reynolds(dp, ep, mu, phi, re, rhog, rhos):
    """

    Calculate minimum fluidization velocity for very small particles where
    Reynolds number < 20 and for very large particles where Reynolds number >
    1000. See Equations 21 and 22 in Chapter 3 of Kunii and Levenspiel [3]_.

    Parameters
    ----------
    dp : float
        Diameter of bed particle [m]
    ep : float
        Void fraction [-]
    mu : float
        Viscosity of gas [kg/ms]
    phi : float
        Sphericity of bed particle [-]
    re : float
        Reynolds number where Re < 20 or Re > 1000 [-]
    rhog : float
        Density of gas [kg/m³]
    rhos : float
        Density of bed particle [kg/m³]

    Returns
    -------
    umf : float
        Minimum fluidization velocity [m/s]

    Example
    -------
    For small Reynolds number where Re = 19

    >>> umf_reynolds(0.0005, 0.46, 3.6e-5, 0.86, 19, 0.44, 2500)
    0.1513

    For large Reynolds number where Re = 1001

    >>> umf_reynolds(0.0005, 0.46, 3.6e-5, 0.86, 1001, 0.44, 2500)
    1.1545

    References
    ----------
    .. [3] Daizo Kunii and Octave Levenspiel. Fluidization Engineering.
       Butterworth-Heinemann, 2nd edition, 1991.
    """

    g = 9.81    # acceleration due to gravity [m/s²]

    if re > 1000:
        umf = np.sqrt((dp * (rhos - rhog) * g * ep**3 * phi) / (1.75 * rhog))
        return umf
    elif re < 20:
        umf = ((dp**2 * (rhos - rhog) * g) / (150 * mu)) * ((ep**3 * phi**2) / (1 - ep))
        return umf
    else:
        raise ValueError('Reynolds number not in applicable range.')



import numpy as np


def dbMoriWen(z, us, umf, d_bed, l_or, dist_type):
    """
    Calculates the equivalent diameter of the gas bubble in the bed, assuming
    that all of the volume in bubbles in the bed were combined into a single
    spherical bubble. This uses the Mori/Wen correlation as given in
    Fluidization Engineering by Kunii & Levenspiel (K/L), eqs. 5.15, 5.19, and
    6.5.

    Parameters
    ----------
    z : float
        Height of the bubble along the vertical axis of the bed [m]
    us : float
        Superficial velocity of the gas [m/s]
    umf : float
        Minimum fluidization velocity [m/s]
    d_bed : float
        Bed diameter [m]
    l_or : float
        Orifice spacing [m] (only if perforated plate distributor used)
    dist_type : string
        Type of distributor plate
        Options are
        'perf_sq' = perforated, square arrangement of orifices
        'perf_tri' = perforated, triangular arrangement of orifices
        'porous' = porous (equivalent to perforated triangular arrangement of
        tiny orifices)

    Returns
    -------
    db : float
        Equivalent bubble diameter at the specified z position [m]
    """

    # Constants
    g_cgs = 981.0    # gravity, cm/s^2

    # Convert all mks units to cgs units for use with the correlation, which
    # appears in K/L in cgs units.
    z_cgs = z * 100.0
    us_cgs = us * 100.0
    umf_cgs = float(umf) * 100.0
    d_bed_cgs = d_bed * 100.0
    l_or_cgs = l_or * 100.0

    # Maximum bubble diameter, cm
    db_max = 0.65 * (np.pi / 4 * d_bed_cgs**2 * (us_cgs - umf_cgs))**0.4

    # Minimum bubble diameter, cm for high flow rate/large bubble sizes at
    # distributor plate. Also works for porous distributors.
    db_min_high = 2.78 / g_cgs * (us_cgs - umf_cgs)**2

    if dist_type == 'perf_sq':

        # Minimum bubble diameter, cm for low flow rate/small bubble sizes at
        # distributor plate
        db_min_low = 1.3 / (g_cgs**0.2) * \
                    ((us_cgs - umf_cgs) * l_or_cgs**2)**0.4

        # Set the minimum bubble diameter based on the orifice spacing
        if db_min_low <= l_or_cgs:
            db_min = db_min_low
        else:
            db_min = db_min_high

    elif dist_type == 'perf_tri':

        # Minimum bubble diameter, cm for low flow rate/small bubble sizes at
        # distributor plate
        db_min_low = 1.3 / (g_cgs**0.2) * ((us_cgs - umf_cgs) *
                                           l_or_cgs**2 * np.sqrt(3)/2)**0.4

        # Set the minimum bubble diameter based on the orifice spacing
        if db_min_low <= l_or_cgs:
            db_min = db_min_low
        else:
            db_min = db_min_high

    elif dist_type == 'porous':

        # Just use the high flow rate equation at the distributor
        db_min = db_min_high

    else:
        raise NotImplementedError(f"Unknown distributor type {dist_type}" +
                                  " in Mori/Wen bubble diameter calculation.")

    # Equivalent bubble diameter, cm
    db = db_max - (db_max - db_min) * np.exp(-0.3 * z_cgs / d_bed_cgs)

    # Constrain to 80% of the diameter of the column
    db = min(0.8*d_bed*100.0, db)

    # Return the bubble diameter, m
    return db / 100.0



def rhog(mw, p, tk):
    """
    Calculate gas density from molecular weight, pressure, and temperature.

    Parameters
    ----------
    mw : float
        Molecular weight of gas [g/mol]
    p : float
        Pressure of the gas [Pa]
    tk : float
        Temperature of the gas [K]

    Returns
    -------
    rho : float
        Density of gas [kg/m^3]

    Examples
    --------
    >>> rhog(28, 101325, 773)
    0.4414
    """
    mw = mw / 1000  # convert g/mol to kg/mol
    r = 8.3145      # ideal gas constant in units of (m^3 Pa)/(K mol)
    rho = (p * mw) / (r * tk)
    return rho




def k_wood(gb, so, x):
    """
    Thermal conductivity of wood based on moisture content, volumetric
    shrinkage, and basic specific gravity

    .. math:: k = G_x (B + C x) + A

    where :math:`k` is thermal conductivity [W/(mK)] of wood, :math:`G_x` is
    specific gravity [-] based on volume at moisture content :math:`x` [%] and
    :math:`A, B, C` are constants.

    The :math:`G_x` term is determined from

    .. math:: G_x = \\frac{G_b}{1 - S_x / 100}

    where :math:`G_b` is basic specific gravity [-] and :math:`S_x` is
    volumetric shrinkage [%] from green condition to moisture content :math:`x`.

    The :math:`S_x` term is calculated from

    .. math:: S_x = S_o \\left(1 - \\frac{x}{MC_{fs}} \\right)

    where :math:`S_o` is volumetric shrinkage [%] from Table 4-3 [1]_ and :math:`MC_{fs}`
    is the fiber saturation point assumed to be 30% moisture content.

    Parameters
    ----------
    gb : float
        Basic specific gravity [-]
    so : float
        Volumetric shrinkage [%]
    x : float
        Moisture content [%]

    Returns
    -------
    k : float
        Thermal conductivity [W/(mK)]

    Example
    -------
    >>> k_wood(0.54, 12.3, 10)
    0.1567

    References
    ----------
    .. [1] Samuel V. Glass and Samuel L. Zelinka. Moisture Relations and
       Physical Properties of Wood. Ch. 4 in Wood Handbook, pp. 1-19, 2010.
    """

    mcfs = 30   # fiber staturation point estimate [%]

    # shrinkage from green to final moisture content, Eq. 4-7 [%]
    sx = so * (1 - x / mcfs)

    # specific gravity based on volume at given moisture content, Eq. 4-9
    gx = gb / (1 - sx / 100)

    # thermal conductivity, Eq. 4-15 [W/(mK)]
    a = 0.01864
    b = 0.1941
    c = 0.004064
    k = gx * (b + c * x) + a

    return k



import setuptools

with open('README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name='chemics',
    version='21.4',
    author='Gavin Wiggins',
    description='A Python package for chemical reactor engineering',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/chemics/chemics',
    packages=setuptools.find_packages(),
    install_requires=['matplotlib', 'numpy', 'pandas', 'scipy'],
    classifiers=(
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ),
    package_data={
        'chemics': ['data/*.csv'],
    },
)



import numpy as np


def disp(D, L, u, t):
    """
    Calculate the 1-D continuum dispersion model for RTD.

    Parameters
    ----------
    D = dispersion coefficient
    L = characteristic mixing length
    u = average fluid velocity
    t = time vector

    Returns
    -------
    et = residence time distribution
    """
    tm1 = u/(np.sqrt(4*np.pi*D*t))
    tm2 = ((L-u*t)**2)/(4*D*t)
    et = tm1*np.exp(-tm2)
    return et


def scstr(n, tau, t):
    """
    Exit age distribution (RTD) for solids in multistaged fluidized beds from
    Eq 5 in Kunii 1991 book from pg 339.

    Parameters
    ----------
    n = number of stages, number of equal-sized beds in series
    tau = total solids residence time, s
    t = time vector, s

    Returns
    -------
    et = exit age distribution or RTD of solids as a whole
    """
    # solids residence time for each stage from Eq. 4
    ti = tau/n
    # RTD of the solids for the beds as a whole from Eq. 5
    et = 1/(np.math.factorial(n-1)*ti)*(t/ti)**(n-1)*np.exp(-t/ti)
    return et


def vusse(n, tau, t, q=1, r=1):
    """
    Residence time distribution function for stirred tank reactor from paper by
    Vusse 1962. Assumes circulating flow, pumping action of stirrer, isotropic
    homogeneous turbulence in the circulating fluid.

    Parameters
    ----------
    n = number of mixing stages, (-)
    tau = circulation time, s
    t = time vector, s
    q = feed rate, m^3/s
    r = circulation rate, m^3/s

    Returns
    -------
    rt = residence time distribution, 1/s

    Example
    -------
    r = rtd(n, tau, t) or rtd(n, tau, t, q=2, r=3)
    """
    a = (n/tau)*(r/(r+q))**(1/n)    # term from Eq. 30
    s = 0                           # initialize summation for g(at) Eq. 31

    for k in range(0, n):
        # summation loop for number of mixing stages, see g(at) from Eq. 31
        tm = np.exp(a*t*(np.cos(2*np.pi*k/n) + 1j*np.sin(2*np.pi*k/n))
                    + 1j*(2*np.pi*k)/n)
        s += tm

    # g as g(at) from Eq. 31 and rt as R(t) from Eq. 30
    g = (1/n)*s
    # vector of complex values for residence time distribution
    rt = (q/(r+q))*((r+q)/r)**((n-1)/n)*(n/tau)*np.exp(-(n*t)/tau)*g
    return rt


def weibull(x, lam, k):
    """
    Weibull distribution function.

    Parameters
    ----------
    x = time parameter
    k = shape parameter
    lam = lambda as scale parameter

    Returns
    -------
    w = weibull distribution
    """
    w = (k/lam)*((x/lam)**(k-1))*np.exp(-(x/lam)**k)
    return w



import pandas as pd
import os


def k_gas_inorganic(formula, temp, full=False):
    """
    Thermal conductivity of gas as a function of temperature. Applicable to gas
    comprised of inorganic compounds. Results based on coefficients from Yaws'
    Critical Property Data for Chemical Engineers and Chemists [1]_.

    Parameters
    ----------
    formula : string
        Molecular formula of the gas.
    temp : float
        Temperature of the gas [K]
    full : bool, optional
        When set to :code:`False` (default) just thermal conductivity is
        returned. When set to :code:`True` then return thermal conductivity and
        other information.

    Returns
    -------
    k_gas : float
        Thermal conductivity of gas [W/(m K)]

    k_gas, cas, tmin, tmax, a, b, c, d : tuple
        Additional values are only returned when keyword :code:`full=True`.

        | k_gas - Thermal conductivity of gas [W/(m K)]
        | cas - CAS number [-]
        | tmin, tmax - Temperature range at which results are applicable [K]
        | a, b, c, d - Values for regression coefficients [-]

    Raises
    ------
    ValueError
        If gas formula is not found in csv data file.
    ValueError
        If gas temperature is not in range between tmin and tmax.

    Examples
    --------
    >>> k_gas_inorganic('N2', 773)
    0.0535

    >>> k_gas_inorganic('N2', 773, full=True)
    (0.0535, '7727-37-9', 63.15, 1500.0, -0.0002267, 0.0001027, -6.0151e-08,
    2.2331e-11)

    References
    ----------
    .. [1] Carl L. Yaws. Table 84. Thermal Conductivity of Gas – Inorganic
       Compounds in Yaws' Critical Property Data for Chemical Engineers and
       Chemists. Published by Knovel, 2014.
    """
    abs_path = os.path.dirname(os.path.abspath(__file__))
    data_file = abs_path + '/data/k-gas-inorganic.csv'

    df = pd.read_csv(data_file, index_col=0)

    if formula not in df.index:
        raise ValueError(f'Gas species {formula} is not available.')

    cas = df.loc[formula]['CAS No.']
    a = df.loc[formula]['A']
    b = df.loc[formula]['B']
    c = df.loc[formula]['C']
    d = df.loc[formula]['D']

    tmin = df.loc[formula]['temperature, Tmin (K)']
    tmax = df.loc[formula]['temperature, Tmax (K)']

    if temp < tmin or temp > tmax:
        raise ValueError('Temperature out of range. Applicable values are '
                         + f'{tmin} - {tmax} K for {formula} gas.')

    k_gas = a + b * temp + c * (temp**2) + d * (temp**3)

    if full:
        return k_gas, cas, tmin, tmax, a, b, c, d
    else:
        return k_gas


def k_gas_organic(formula, temp, cas=None, full=False):
    """
    Thermal conductivity of gas as a function of temperature. Applicable to gas
    comprised of organic compounds. Results based on coefficients from Yaws'
    Critical Property Data for Chemical Engineers and Chemists [2]_.

    Parameters
    ----------
    formula : string
        Molecular formula of the gas.
    temp : float
        Temperature of the gas [K]
    cas : string
        CAS number of the gas, required for some species [-]
    full : bool, optional
        When set to :code:`False` (default) just thermal conductivity is
        returned. When set to :code:`True` then return thermal conductivity and
        other information.

    Returns
    -------
    k_gas : float
        Thermal conductivity of gas [W/(m K)]

    k_gas, cas, tmin, tmax, a, b, c, d : tuple
        Additional values are only returned when keyword :code:`full=True`.

        | k_gas - Thermal conductivity of gas [W/(m K)]
        | cas - CAS number [-]
        | tmin, tmax - Temperature range at which results are applicable [K]
        | a, b, c, d - Values for regression coefficients [-]

    Raises
    ------
    ValueError
        If gas formula is not available in CSV data file.
    ValueError
        If multiple substances have same formula. Require CAS number.
    ValueError
        If gas temperature is not in range between Tmin and Tmax.

    Examples
    --------
    >>> k_gas_organic('CO', 801)
    0.05722

    >>> k_gas_organic('C18H38O', 920, cas='593-32-8')
    0.04174

    References
    ----------
    .. [2] Carl L. Yaws. Table 85. Thermal Conductivity of Gas – Organic
       Compounds in Yaws' Critical Property Data for Chemical Engineers and
       Chemists. Published by Knovel, 2014.
    """
    abs_path = os.path.dirname(os.path.abspath(__file__))
    data_file = abs_path + '/data/k-gas-organic.csv'

    df = pd.read_csv(data_file, index_col=0)

    if formula not in df.index:
        raise ValueError(f'Gas species {formula} is not available.')

    if isinstance(df.loc[formula], pd.DataFrame) and cas is None:
        raise ValueError(f'Multiple substances available for {formula}.'
                         + ' Include CAS number as a function parameter.')

    if cas:
        df = df[df['CAS No.'] == str(cas)]

    cas = df.loc[formula]['CAS No.']
    tmin = df.loc[formula]['temperature, Tmin (K)']
    tmax = df.loc[formula]['temperature, Tmax (K)']
    a = df.loc[formula]['A']
    b = df.loc[formula]['B']
    c = df.loc[formula]['C']
    d = df.loc[formula]['D']

    if temp < tmin or temp > tmax:
        raise ValueError('Temperature out of range. Applicable values are '
                         + f'{tmin} - {tmax} K for {formula} gas.')

    k_gas = a + b * temp + c * (temp**2) + d * (temp**3)

    if full:
        return k_gas, cas, tmin, tmax, a, b, c, d
    else:
        return k_gas



import numpy as np


def ultimate_bases(c, h, o, n, s, ash, moisture, disp=False):
    """
    Convert ultimate analysis from as-received basis (% ar) to dry basis (%
    dry) and dry ash-free (% daf) basis.

    Parameters
    ----------
    c : float
        Percent carbon
    h : float
        Percent hydrogen
    o : float
        Percent oxygen
    n : float
        Percent nitrogen
    s : float
        Percent sulfur
    ash : float
        Percent ash
    moisture : float
        Percent moisture
    disp : bool, optional
        Display results to console, default is `False`

    Returns
    -------
    bases : dict
        Ultimate analysis bases calculated from given as-received basis. Bases
        are represented by keys in the dictionary where `'ar'` is as-received
        basis with values that represent `[C, H, O, N, S, ash, moisture]`,
        `'dry'` is dry basis with values that represent `[C, H, O, N, S,
        ash]`, and `'daf'` is dry ash-free basis with values of `[C, H, O, N,
        S]`.

    Raises
    ------
    ValueError
        If the ultimate analysis sum is not 100.

    Example
    -------
    >>> ultimate_bases(49.52, 5.28, 38.35, 0.15, 0.02, 0.64, 6.04)
    {
        'ar': [49.52, 5.28, 38.35, 0.15, 0.02, 0.64, 6.04],
        'dry': [52.70, 5.61, 40.81, 0.15, 0.02, 0.68],
        'daf': [53.06, 5.65, 41.09, 0.16, 0.02]
    }
    """

    # as-received basis (% ar)
    ult_ar = [c, h, o, n, s, ash, moisture]
    sum_ar = sum(ult_ar)

    # make sure ultimate analysis sums to 100
    if not np.isclose(sum_ar, 100.0):
        raise ValueError('Sum of ultimate analysis values must be 100')

    # dry basis (% dry)
    ult_dry = [100 * x / (sum_ar - ult_ar[-1]) for x in ult_ar[:-1]]
    sum_dry = sum(ult_dry)

    # dry ash-free basis (% daf)
    ult_daf = [100 * x / (sum_dry - ult_dry[-1]) for x in ult_dry[:-1]]
    sum_daf = sum(ult_daf)

    # print results to console if `disp=True`
    if disp:
        results = (
            f'               % ar    % dry    % daf\n'
            f'C          {ult_ar[0]:8} {ult_dry[0]:8.2f} {ult_daf[0]:8.2f}\n'
            f'H          {ult_ar[1]:8} {ult_dry[1]:8.2f} {ult_daf[1]:8.2f}\n'
            f'O          {ult_ar[2]:8} {ult_dry[2]:8.2f} {ult_daf[2]:8.2f}\n'
            f'N          {ult_ar[3]:8} {ult_dry[3]:8.2f} {ult_daf[3]:8.2f}\n'
            f'S          {ult_ar[4]:8} {ult_dry[4]:8.2f} {ult_daf[4]:8.2f}\n'
            f'ash        {ult_ar[5]:8} {ult_dry[5]:8.2f}\n'
            f'moisture   {ult_ar[6]:8}\n'
            f'sum        {sum_ar:8.2f} {sum_dry:8.2f} {sum_daf:8.2f}\n'
        )
        print(results)

    # results dictionary
    bases = {
        'ar': ult_ar,
        'dry': ult_dry,
        'daf': ult_daf
    }

    return bases



import numpy as np
import re
from .atomic_elements import atomic_elements

__all__ = ['mw', 'mw_mix']


def _find_end(tokens):
    """
    Find index of closing parenthesis. Based on function from Rhomboid at
    https://gist.github.com/Rhomboid/5994999.

    Parameters
    ----------
    tokens : list of str
        List of strings representing molecular formula.

    Return
    ------
    idx : int
        Index of closing parenthesis.
    """
    count = 0

    for idx, t in enumerate(tokens):
        if t == ')':
            count -= 1
            if count == 0:
                return idx
        elif t == '(':
            count += 1

    raise ValueError('Unmatched parentheses')


def _parse(tokens, stack):
    """
    Parse items in formula list. Get atomic weight of each item or multiply by
    number of elements. Final value is molecular weight of the formula. Based
    on function from Rhomboid at https://gist.github.com/Rhomboid/5994999.

    Parameters
    ----------
    tokens : list of str
        List of strings representing molecular formula.
    stack : list of float
        Molecular weights of each element in formula.

    Returns
    -------
    mole_weight : float
        Sum of molecular weights from molecular formula.
    """
    if len(tokens) == 0:
        mole_weight = sum(stack)
        return mole_weight

    t = tokens[0]

    if t == '(':
        end = _find_end(tokens)
        stack.append(_parse(tokens[1:end], []))
        return _parse(tokens[end + 1:], stack)
    elif t in atomic_elements:
        stack.append(atomic_elements[t]['atomic_weight'])
    elif t.isdigit():
        stack[-1] *= int(t)

    return _parse(tokens[1:], stack)


def mw(formula):
    """
    Tokenize a molecular formula to determine total molecular weight.
    Calculation is based on atomic weight values from IUPAC [1]_.

    Parameters
    ----------
    formula : str
        Molecular formula or element.

    Returns
    -------
    mw : float
       Molecular weight of the formula or element [g/mol]

    Examples
    --------
    >>> mw('C')
    12.011

    >>> mw('CH4')
    16.04

    >>> mw('(NH4)2SO4')
    132.13

    References
    ----------
    .. [1] IUPAC Periodic Table of the Elements. International Union of Pure
       and Applied Chemistry, 2016.
       https://iupac.org/what-we-do/periodic-table-of-elements/.
    """
    tokens = re.findall(r'[A-Z][a-z]*|\d+|\(|\)', formula)
    molecular_weight = _parse(tokens, [])
    return molecular_weight


def mw_mix(mws, xs):
    """
    Molecular weight of a gas mixture calculated as a weighted mean.

    Parameters
    ----------
    mws : list, tuple, or array
        Molecular weight of each gas component [g/mol]
    xs : list, tuple, or array
        Mole fraction of each gas component [-]

    Returns
    -------
    mw_mix : float
        Molecular weight of a gas mixture [g/mol]

    Raises
    ------
    ValueError
        If sum of mole fractions does not equal 1.0

    Examples
    --------
    >>> mw_h2 = cm.mw('H2')
    ... mw_n2 = cm.mw('N2')
    ... mw_mix([mw_h2, mw_n2], [0.8, 0.2])
    7.2156

    >>> mw_h2 = cm.mw('H2')
    ... mw_n2 = cm.mw('N2')
    ... mw_ch4 = cm.mw('CH4')
    ... mw_mix([mw_h2, mw_n2, mw_ch4], [0.4, 0.1, 0.5])
    11.6293
    """
    mws = np.asarray(mws)
    xs = np.asarray(xs)
    if not np.isclose(xs.sum(), 1.0):
        raise ValueError('Sum of mole fractions must be 1.0')
    molwt_mix = np.average(mws, weights=xs)
    return molwt_mix




def fbexp(db, dp, rhog, rhos, umf, us):
    """
    Bed expansion factor for calculating expanded bed height of a bubbling
    fluidized bed reactor. See equations 14.7 and 14.8 in Souza-Santos [1]_.

    Parameters
    ----------
    db : float
        Diameter of the bed [m]
    dp : float
        Diameter of the bed particle [m]
    rhog : float
        Density of gas [kg/m^3]
    rhos : float
        Density of bed particle [kg/m^3]
    umf : float
        Minimum fluidization velocity [m/s]
    us : float
        Superficial gas velocity [m/s]

    Returns
    -------
    fbx : float
        Bed expansion factor [-]

    Example
    -------
    >>> umf = 0.1157
    ... us = 3.0*umf
    ... fbexp(0.05232, 0.0004, 0.4413, 2500, 0.1157, us)
    1.4864

    References
    ----------
    .. [1] Marcio de Souza-Santos. Solid Fuels Combustion and Gasification:
       Modeling, Simulation, and Equipment Operations. CRC Press, Taylor and
       Francis Group, 2nd edition, 2010.
    """

    if db < 0.0635:
        # diameter of bed as db < 0.0635 m from Eq 14.7
        tm1 = 1.032 * ((us - umf)**0.57) * (rhog**0.083)
        tm2 = (rhos**0.166) * (umf**0.063) * (db**0.445)
        fbx = 1 + (tm1 / tm2)
    else:
        # diameter of bed as db >= 0.0635 m from Eq 14.8
        tm1 = 14.314 * ((us - umf)**0.738) * (dp**1.006) * (rhos**0.376)
        tm2 = (rhog**0.126) * (umf**0.937)
        fbx = 1 + (tm1 / tm2)

    return fbx
